# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the debian pipeline workflow."""

from typing import Any, ClassVar, override

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianSourcePackage,
    TaskTypes,
)
from debusine.assets import KeyPurpose
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, WorkRequest
from debusine.server.workflows import (
    DebianPipelineWorkflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.models import (
    DebianPipelineWorkflowData,
    DebianPipelineWorkflowDynamicData,
    QAFailOn,
    QAFailurePolicy,
    RegressionTrackingUpdateQAResults,
)
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    BackendType,
    CollectionItemMatcherKind,
    ExtraDebusineRepository,
    ExtraExternalRepository,
    InputArtifactSingle,
    LintianFailOnSeverity,
    OutputData,
    OutputDataError,
)


class DebianPipelineWorkflowTests(WorkflowTestCase[DebianPipelineWorkflow]):
    """Unit tests for :py:class:`DebianPipelineWorkflow`."""

    SAMPLE_TASK_DATA = {
        "vendor": "debian",
        "codename": "sid",
    }

    source_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.id

    def create_debian_pipeline_workflow(
        self,
        *,
        extra_task_data: dict[str, Any] | None = None,
        validate: bool = True,
    ) -> DebianPipelineWorkflow:
        """Create a debian pipeline workflow."""
        task_data = {
            "source_artifact": self.source_artifact.id,
            "vendor": "debian",
            "codename": "bookworm",
        }
        if extra_task_data is not None:
            task_data.update(extra_task_data)
        wr = self.playground.create_workflow(
            task_name="debian_pipeline", task_data=task_data, validate=validate
        )
        return self.get_workflow(wr)

    def orchestrate(
        self, extra_data: dict[str, Any], expect_success: bool = True
    ) -> WorkRequest:
        """Create and orchestrate a DebianPipelineWorkflow."""
        data = {
            "source_artifact": self.source_artifact.id,
            "vendor": "debian",
            "codename": "bookworm",
            "enable_autopkgtest": False,
            "enable_lintian": False,
            "enable_piuparts": False,
            "enable_blhc": False,
            **extra_data,
        }

        wr = self.playground.create_workflow(
            task_name="debian_pipeline",
            task_data=DebianPipelineWorkflowData(**data),
        )
        self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(
            self.playground.orchestrate_workflow(wr, use_data=False),
            expect_success,
        )

        return wr

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        workflow = self.create_debian_pipeline_workflow()

        workflow.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        workflow = self.create_debian_pipeline_workflow(
            extra_task_data={
                "qa_suite": "nonexistent@debian:suite",
                "enable_reverse_dependencies_autopkgtest": True,
            },
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            workflow.validate_input()

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data includes a parameter summary."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebianPipelineWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                input_source_artifact_id=self.source_artifact.id,
            ),
        )

    def test_compute_dynamic_data_source_artifact_wrong_category(self) -> None:
        """`source_artifact` must be a source-package or upload."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"^source_artifact: unexpected artifact category: 'debusine:test'. "
            r"Valid categories: \['debian:source-package', 'debian:upload'\]$",
        ):
            self.playground.create_workflow(
                task_name="debian_pipeline",
                task_data=DebianPipelineWorkflowData(
                    source_artifact=artifact.id, vendor="debian", codename="sid"
                ),
            )

    def test_get_input_artifacts(self) -> None:
        upload_artifacts = self.playground.create_upload_artifacts()
        self.task_data["source_artifact"] = upload_artifacts.upload.id
        workflow = self.task

        # workflow.dynamic_data is None
        self.assertIsNone(workflow.dynamic_data)
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=None,
                )
            ],
        )

        # workflow.dynamic_data.source_artifact_id is None
        workflow.dynamic_data = DebianPipelineWorkflowDynamicData()
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=None,
                )
            ],
        )

        # workflow.dynamic_data.source_artifact_id is the id: return it
        workflow.dynamic_data.input_source_artifact_id = (
            upload_artifacts.upload.id
        )
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=upload_artifacts.upload.id,
                )
            ],
        )

    def test_get_extra_repositories_none(self) -> None:
        self.assertIsNone(self.task.get_extra_repositories())

    def test_get_extra_repositories_explicit(self) -> None:
        external = ExtraExternalRepository(
            url="http://example.com/", suite="sid", components=["main"]
        )
        self.task_data["extra_repositories"] = [external]
        self.assertEqual(self.task.get_extra_repositories(), [external])

    def test_get_extra_repositories_suite_without_active_indexes(self) -> None:
        self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        self.task_data["suite"] = f"sid@{CollectionCategory.SUITE}"
        self.assertIsNone(self.task.get_extra_repositories())

    def test_get_extra_repositories_suite(self) -> None:
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        sid.manager.add_artifact(
            self.playground.create_repository_index("InRelease"),
            user=self.playground.get_default_user(),
            variables={"path": "InRelease"},
        )
        self.task_data["suite"] = f"sid@{CollectionCategory.SUITE}"
        self.assertEqual(
            self.task.get_extra_repositories(),
            [ExtraDebusineRepository(suite=f"sid@{CollectionCategory.SUITE}")],
        )

    def test_get_extra_repositories_suite_and_external(self) -> None:
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        sid.manager.add_artifact(
            self.playground.create_repository_index("InRelease"),
            user=self.playground.get_default_user(),
            variables={"path": "InRelease"},
        )
        external = ExtraExternalRepository(
            url="http://example.com/", suite="sid", components=["main"]
        )
        self.task_data["suite"] = f"sid@{CollectionCategory.SUITE}"
        self.task_data["extra_repositories"] = [external]
        self.assertEqual(
            self.task.get_extra_repositories(),
            [
                ExtraDebusineRepository(
                    suite=f"sid@{CollectionCategory.SUITE}"
                ),
                external,
            ],
        )

    def test_source_package_upload(self) -> None:
        """`source_artifact` can be an upload."""
        upload_artifacts = self.playground.create_upload_artifacts()
        self.task_data["source_artifact"] = upload_artifacts.upload.id
        self.assertEqual(
            workflow_utils.source_package(self.task), upload_artifacts.source
        )

    def test_populate_use_available_architectures(self) -> None:
        """
        Test populate use available architectures.

        The user didn't specify "architectures", DebianPipelineWorkflow
        checks available architectures and "all" and use them.
        """
        collection = self.playground.create_collection(
            "debian",
            CollectionCategory.ENVIRONMENTS,
            workspace=self.playground.get_default_workspace(),
        )
        for arch in ["amd64", "i386"]:
            artifact, _ = self.playground.create_artifact(
                category=ArtifactCategory.SYSTEM_TARBALL,
                data={"codename": "bookworm", "architecture": arch},
            )
            collection.manager.add_artifact(
                artifact, user=self.playground.get_default_user()
            )

        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "vendor": "debian",
                "codename": "bookworm",
            }
        )

        sbuild = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            sbuild.task_data["architectures"], ["all", "amd64", "i386"]
        )

    def test_populate_sbuild(self) -> None:
        """Test populate create sbuild."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_blhc": False,
            }
        )

        sbuild = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(sbuild.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            sbuild.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "backend": BackendType.AUTO,
                "environment_variant": None,
                "extra_repositories": None,
                "input": {"source_artifact": self.source_artifact.id},
                "target_distribution": "debian:bookworm",
                "signing_template_names": {},
            },
        )
        self.assertEqual(
            sbuild.workflow_data_json,
            {"display_name": "sbuild", "step": "sbuild"},
        )

        # SbuildWorkflow.populate() was called and created its tasks
        self.assertTrue(sbuild.children.exists())
        for child in sbuild.children.all():
            self.assertEqual(child.status, WorkRequest.Statuses.PENDING)

        self.assertFalse(
            workflow.children.filter(
                task_name="qa", task_type=TaskTypes.WORKFLOW
            ).exists()
        )

    def test_populate_sbuild_with_suite(self) -> None:
        bookworm = self.playground.create_collection(
            name="bookworm", category=CollectionCategory.SUITE
        )
        bookworm.manager.add_artifact(
            self.playground.create_repository_index("InRelease"),
            user=self.playground.get_default_user(),
            variables={"path": "InRelease"},
        )

        workflow = self.orchestrate(
            extra_data={
                "suite": f"bookworm@{CollectionCategory.SUITE}",
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
            }
        )

        sbuild = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(sbuild.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            sbuild.task_data["extra_repositories"],
            [{"suite": f"bookworm@{CollectionCategory.SUITE}"}],
        )

    def assert_qa(
        self, workflow: WorkRequest, task_data: dict[str, Any]
    ) -> None:
        """Assert workflow has a sub-workflow qa with task_data."""
        qa = workflow.children.get(task_name="qa", task_type=TaskTypes.WORKFLOW)
        self.assertEqual(qa.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(qa.task_data, task_data)

        self.assertEqual(
            qa.workflow_data_json, {"display_name": "QA", "step": "qa"}
        )

        # QaWorkflow.populate() was called and created its tasks
        self.assertTrue(qa.children.exists())
        for child in qa.children.all():
            # Children are sub-workflows, so go straight to RUNNING.
            self.assertEqual(child.status, WorkRequest.Statuses.RUNNING)

        self.assertQuerySetEqual(
            qa.children.earliest("id").dependencies.all(), []
        )

    def test_populate_qa_with_sbuild_subset(self) -> None:
        """`sbuild` may only create a subset of the requested architectures."""
        self.assertEqual(
            self.source_artifact.data["dsc_fields"]["Architecture"], "any"
        )

        workflow = self.orchestrate(
            extra_data={
                "architectures": ["all", "amd64"],
                "source_artifact": f"{self.source_artifact.id}@artifacts",
                "enable_lintian": True,
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
            }
        )

        sbuild = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            sbuild.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["all", "amd64"],
                "backend": BackendType.AUTO,
                "environment_variant": None,
                "input": {
                    "source_artifact": f"{self.source_artifact.id}@artifacts"
                },
                "target_distribution": "debian:bookworm",
                "signing_template_names": {},
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
            },
        )
        self.assertEqual(
            sbuild.workflow_data_json,
            {"display_name": "sbuild", "step": "sbuild"},
        )

        self.assertTrue(sbuild.children.exists())
        self.assert_qa(
            workflow,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["all", "amd64"],
                "autopkgtest_backend": BackendType.AUTO,
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "bookworm",
                "extra_repositories": [
                    {
                        "url": "http://example.com/",
                        "suite": "bookworm",
                        "components": ["main"],
                    }
                ],
                "enable_autopkgtest": False,
                "enable_lintian": True,
                "enable_piuparts": False,
                "enable_regression_tracking": False,
                "enable_reverse_dependencies_autopkgtest": False,
                "enable_debdiff": False,
                "enable_blhc": False,
                "fail_on": QAFailOn.FAILURE,
                "lintian_backend": BackendType.AUTO,
                "lintian_fail_on_severity": LintianFailOnSeverity.ERROR,
                "piuparts_backend": BackendType.AUTO,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": f"{self.source_artifact.id}@artifacts",
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )

    def test_populate_qa_lintian(self) -> None:
        """Test populate create qa: lintian enabled."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_lintian": True,
            }
        )
        self.assert_qa(
            workflow,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "autopkgtest_backend": BackendType.AUTO,
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "bookworm",
                "extra_repositories": None,
                "enable_autopkgtest": False,
                "enable_lintian": True,
                "enable_piuparts": False,
                "enable_regression_tracking": False,
                "enable_reverse_dependencies_autopkgtest": False,
                "enable_debdiff": False,
                "enable_blhc": False,
                "fail_on": QAFailOn.FAILURE,
                "lintian_backend": BackendType.AUTO,
                "lintian_fail_on_severity": LintianFailOnSeverity.ERROR,
                "piuparts_backend": BackendType.AUTO,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": self.source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )

    def test_populate_qa_autopkgtest(self) -> None:
        """Test populate create qa: autopkgtest enabled."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_autopkgtest": True,
            }
        )

        self.assert_qa(
            workflow,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "autopkgtest_backend": BackendType.AUTO,
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "bookworm",
                "extra_repositories": None,
                "enable_autopkgtest": True,
                "enable_lintian": False,
                "enable_piuparts": False,
                "enable_regression_tracking": False,
                "enable_reverse_dependencies_autopkgtest": False,
                "enable_debdiff": False,
                "enable_blhc": False,
                "fail_on": QAFailOn.FAILURE,
                "lintian_backend": BackendType.AUTO,
                "lintian_fail_on_severity": LintianFailOnSeverity.ERROR,
                "piuparts_backend": BackendType.AUTO,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": self.source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )

    def test_populate_qa_piuparts(self) -> None:
        """Test populate create qa: piuparts enabled."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_piuparts": True,
            }
        )

        self.assert_qa(
            workflow,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "autopkgtest_backend": BackendType.AUTO,
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "bookworm",
                "extra_repositories": None,
                "enable_autopkgtest": False,
                "enable_lintian": False,
                "enable_piuparts": True,
                "enable_regression_tracking": False,
                "enable_reverse_dependencies_autopkgtest": False,
                "enable_debdiff": False,
                "enable_blhc": False,
                "fail_on": QAFailOn.FAILURE,
                "lintian_backend": BackendType.AUTO,
                "lintian_fail_on_severity": LintianFailOnSeverity.ERROR,
                "piuparts_backend": BackendType.AUTO,
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": self.source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )

    def test_populate_qa_debdiff(self) -> None:
        """Test populate create qa: debdiff enabled."""
        trixie = self.playground.create_collection(
            name="trixie",
            category=CollectionCategory.SUITE,
        )

        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_debdiff": True,
                "qa_suite": "trixie@debian:suite",
            }
        )

        qa = workflow.children.get(task_name="qa", task_type=TaskTypes.WORKFLOW)

        self.assertEqual(
            qa.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "autopkgtest_backend": BackendType.AUTO,
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "bookworm",
                "extra_repositories": None,
                "enable_autopkgtest": False,
                "enable_lintian": False,
                "enable_piuparts": False,
                "enable_regression_tracking": False,
                "enable_reverse_dependencies_autopkgtest": False,
                "enable_debdiff": True,
                "enable_blhc": False,
                "fail_on": QAFailOn.FAILURE,
                "lintian_backend": BackendType.AUTO,
                "lintian_fail_on_severity": LintianFailOnSeverity.ERROR,
                "piuparts_backend": BackendType.AUTO,
                "prefix": "",
                "qa_suite": str(trixie),
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": self.source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )

        self.assertEqual(qa.status, WorkRequest.Statuses.RUNNING)

        self.assertEqual(
            qa.children.filter(
                task_name="debdiff", task_type=TaskTypes.WORKFLOW
            ).count(),
            1,
        )

    def test_populate_qa_piuparts_configuration(self) -> None:
        """Test populate create qa: piuparts non-default configuration."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_piuparts": True,
                "piuparts_backend": BackendType.INCUS_LXC,
                "piuparts_environment": "debian/match:codename=trixie",
            }
        )

        self.assert_qa(
            workflow,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["amd64"],
                "autopkgtest_backend": BackendType.AUTO,
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "bookworm",
                "extra_repositories": None,
                "enable_autopkgtest": False,
                "enable_lintian": False,
                "enable_piuparts": True,
                "enable_regression_tracking": False,
                "enable_reverse_dependencies_autopkgtest": False,
                "enable_debdiff": False,
                "enable_blhc": False,
                "fail_on": QAFailOn.FAILURE,
                "lintian_backend": BackendType.AUTO,
                "lintian_fail_on_severity": LintianFailOnSeverity.ERROR,
                "piuparts_backend": BackendType.INCUS_LXC,
                "piuparts_environment": "debian/match:codename=trixie",
                "prefix": "",
                "qa_suite": None,
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": self.source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
            },
        )

    def test_populate_qa_enable_regression_tracking_no_reference_source(
        self,
    ) -> None:
        self.playground.create_collection(
            name="trixie", category=CollectionCategory.SUITE
        )
        self.playground.create_collection(
            "trixie", CollectionCategory.QA_RESULTS
        )

        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "qa_suite": f"trixie@{CollectionCategory.SUITE}",
                "regression_tracking_qa_results": (
                    f"trixie@{CollectionCategory.QA_RESULTS}"
                ),
                "enable_regression_tracking": True,
                "enable_autopkgtest": True,
            }
        )

        normal_qa = workflow.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="qa",
            task_data__update_qa_results=RegressionTrackingUpdateQAResults.NO,
        )
        self.assertDictContainsAll(
            normal_qa.task_data,
            {
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "enable_regression_tracking": False,
                "fail_on": QAFailOn.FAILURE,
                "prefix": "",
                "reference_prefix": "",
                "reference_qa_results": None,
                "reference_source_artifact": None,
                "source_artifact": self.source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            normal_qa.workflow_data_json, {"display_name": "QA", "step": "qa"}
        )

        self.assertFalse(
            workflow.children.filter(
                task_type=TaskTypes.WORKFLOW,
                task_name="qa",
                task_data__update_qa_results=(
                    RegressionTrackingUpdateQAResults.YES
                ),
            ).exists()
        )

    def test_populate_qa_enable_regression_tracking(self) -> None:
        reference_source_artifact = self.source_artifact
        reference_binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                package="hello",
                version="1.0-1",
            )
        )
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )
        trixie = self.playground.create_collection(
            name="trixie", category=CollectionCategory.SUITE
        )
        trixie.manager.add_artifact(
            reference_source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        trixie.manager.add_artifact(
            reference_binary_artifact,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        self.playground.create_collection(
            "trixie", CollectionCategory.QA_RESULTS
        )

        workflow = self.orchestrate(
            extra_data={
                "architectures": ["all", "amd64"],
                "source_artifact": new_source_artifact.id,
                "qa_suite": f"trixie@{CollectionCategory.SUITE}",
                "regression_tracking_qa_results": (
                    f"trixie@{CollectionCategory.QA_RESULTS}"
                ),
                "enable_regression_tracking": True,
                "enable_autopkgtest": True,
                "enable_debdiff": True,
            }
        )

        normal_qa = workflow.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="qa",
            task_data__update_qa_results=RegressionTrackingUpdateQAResults.NO,
        )
        self.assertDictContainsAll(
            normal_qa.task_data,
            {
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "enable_debdiff": True,
                "enable_regression_tracking": True,
                "fail_on": QAFailOn.REGRESSION,
                "prefix": "",
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": (
                    f"trixie@{CollectionCategory.QA_RESULTS}"
                ),
                "reference_source_artifact": reference_source_artifact.id,
                "source_artifact": new_source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
            },
        )
        self.assertEqual(
            normal_qa.workflow_data_json, {"display_name": "QA", "step": "qa"}
        )

        extra_qa = workflow.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="qa",
            task_data__update_qa_results=RegressionTrackingUpdateQAResults.YES,
        )
        self.assertDictContainsAll(
            extra_qa.task_data,
            {
                "binary_artifacts": [
                    {
                        "category": ArtifactCategory.BINARY_PACKAGE,
                        "collection": f"trixie@{CollectionCategory.SUITE}",
                        "data_matchers": [
                            [
                                "srcpkg_name",
                                {
                                    "kind": CollectionItemMatcherKind.EXACT,
                                    "value": "hello",
                                },
                            ],
                            [
                                "srcpkg_version",
                                {
                                    "kind": CollectionItemMatcherKind.EXACT,
                                    "value": "1.0-1",
                                },
                            ],
                        ],
                    }
                ],
                "enable_debdiff": False,
                "enable_regression_tracking": False,
                "fail_on": QAFailOn.NEVER,
                "prefix": "reference-qa-result|",
                "reference_prefix": "",
                "reference_qa_results": (
                    f"trixie@{CollectionCategory.QA_RESULTS}"
                ),
                "reference_source_artifact": None,
                "source_artifact": reference_source_artifact.id,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
            },
        )
        self.assertEqual(
            extra_qa.workflow_data_json,
            {
                "display_name": "QA (reference tests for hello_1.0-1)",
                "step": "qa-reference",
                "expanded": False,
            },
        )

        normal_autopkgtest = normal_qa.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        normal_autopkgtest_regression_analysis = (
            normal_autopkgtest.children.workflow_callbacks().get(
                workflow_data_json__step="regression-analysis"
            )
        )
        extra_autopkgtest = extra_qa.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        self.assertQuerySetEqual(
            normal_autopkgtest_regression_analysis.dependencies.all(),
            [
                normal_autopkgtest.children.get(
                    task_type=TaskTypes.WORKER, task_name="autopkgtest"
                ),
                extra_autopkgtest.children.get(
                    task_type=TaskTypes.WORKER, task_name="autopkgtest"
                ),
            ],
            ordered=False,
        )

    def test_populate_qa_with_suite(self) -> None:
        bookworm = self.playground.create_collection(
            name="bookworm", category=CollectionCategory.SUITE
        )
        bookworm.manager.add_artifact(
            self.playground.create_repository_index("InRelease"),
            user=self.playground.get_default_user(),
            variables={"path": "InRelease"},
        )

        workflow = self.orchestrate(
            extra_data={
                "suite": f"bookworm@{CollectionCategory.SUITE}",
                "architectures": ["all", "amd64"],
                "source_artifact": f"{self.source_artifact.id}@artifacts",
                "enable_lintian": True,
            }
        )

        qa = workflow.children.get(task_name="qa", task_type=TaskTypes.WORKFLOW)
        self.assertEqual(qa.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(
            qa.task_data["extra_repositories"],
            [{"suite": f"bookworm@{CollectionCategory.SUITE}"}],
        )

    def test_populate_architectures_allow_deny(self) -> None:
        """Populate uses architectures, architectures_allow/deny."""
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "architectures": ["amd64", "i386", "arm64"],
                "architectures_allowlist": ["amd64", "i386"],
                "architectures_denylist": ["amd64"],
            }
        )

        sbuild = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(
            sbuild.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "architectures": ["i386"],
                "backend": BackendType.AUTO,
                "environment_variant": None,
                "extra_repositories": None,
                "input": {"source_artifact": self.source_artifact.id},
                "target_distribution": "debian:bookworm",
                "signing_template_names": {},
            },
        )

    def test_populate_make_signed_source_missing_signed_source_purpose(
        self,
    ) -> None:
        """Orchestrator error: missing required field."""
        msg = (
            'Orchestrator failed: "make_signed_source_purpose" must be '
            'set when signing the source'
        )
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "architectures": ["amd64"],
                "enable_make_signed_source": True,
                "signing_template_names": {"amd64": ["hello"]},
            },
            expect_success=False,
        )
        self.assertEqual(
            workflow.output_data,
            OutputData(
                errors=[
                    OutputDataError(message=msg, code="orchestrator-failed")
                ]
            ),
        )

    def test_populate_make_signed_source_missing_make_signed_source_key(
        self,
    ) -> None:
        """Orchestrator error: missing required field."""
        msg = (
            'Orchestrator failed: "make_signed_source_key" must be set '
            'when signing the source'
        )
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "architectures": ["amd64"],
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {"amd64": ["hello"]},
            },
            expect_success=False,
        )
        self.assertEqual(
            workflow.output_data,
            OutputData(
                errors=[
                    OutputDataError(message=msg, code="orchestrator-failed")
                ]
            ),
        )

    def test_populate_make_signed_source(self) -> None:
        """Populate create a make_signed_source workflow."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {"amd64": ["hello"]},
                "make_signed_source_key": "ABC123",
            }
        )

        expected_binary_artifacts = [
            {
                "collection": "internal@collections",
                "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                "name_matcher": {
                    "kind": CollectionItemMatcherKind.STARTSWITH,
                    "value": "build-",
                },
            }
        ]
        make_signed_source = workflow.children.get(
            task_name="make_signed_source", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(
            make_signed_source.task_data,
            {
                "architectures": ["amd64"],
                "binary_artifacts": expected_binary_artifacts,
                "codename": "bookworm",
                "key": "ABC123",
                "purpose": "uefi",
                "sbuild_backend": BackendType.AUTO,
                "signing_template_artifacts": [
                    "internal@collections/name:signing-template-amd64-hello"
                ],
                "vendor": "debian",
            },
        )

        self.assertEqual(
            make_signed_source.workflow_data_json,
            {
                "display_name": "make signed source",
                "step": "make_signed_source",
            },
        )
        self.assertQuerySetEqual(make_signed_source.dependencies.all(), [])

        # MakeSignedSource.populate() was called and created its tasks.
        # They are all blocked, with the first one depending on sbuild and
        # the others depending on the previous one in sequence.
        self.assertQuerySetEqual(
            make_signed_source.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "extractforsigning",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                ),
                (
                    TaskTypes.SIGNING,
                    "sign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "extractforsigning",
                ),
                (
                    TaskTypes.WORKER,
                    "assemblesignedsource",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.SIGNING,
                    "sign",
                ),
                (
                    TaskTypes.WORKFLOW,
                    "sbuild",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "assemblesignedsource",
                ),
            ],
        )

    def test_populate_make_signed_source_qa_failure_policy_fail(self) -> None:
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_autopkgtest": True,
                "qa_failure_policy": QAFailurePolicy.FAIL,
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {"amd64": ["hello"]},
                "make_signed_source_key": "ABC123",
            }
        )

        self.assertFalse(
            workflow.children.filter(
                task_type=TaskTypes.WAIT, task_name="confirm"
            ).exists()
        )

        qa = workflow.children.get(task_type=TaskTypes.WORKFLOW, task_name="qa")
        make_signed_source = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="make_signed_source"
        )
        self.assertQuerySetEqual(make_signed_source.dependencies.all(), [qa])

    def test_populate_make_signed_source_qa_failure_policy_confirm(
        self,
    ) -> None:
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_autopkgtest": True,
                "qa_failure_policy": QAFailurePolicy.CONFIRM,
                "enable_confirmation": True,
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {"amd64": ["hello"]},
                "make_signed_source_key": "ABC123",
            }
        )

        qa = workflow.children.get(task_type=TaskTypes.WORKFLOW, task_name="qa")
        confirm = workflow.children.get(
            task_type=TaskTypes.WAIT, task_name="confirm"
        )
        self.assertEqual(
            confirm.task_data,
            {
                "auto_confirm_if_no_failure": True,
                "confirm_label": "Continue the workflow despite QA failures",
            },
        )
        self.assertEqual(
            confirm.workflow_data_json,
            {"display_name": "Confirm", "step": "confirm", "needs_input": True},
        )
        self.assertQuerySetEqual(confirm.dependencies.all(), [qa])

        make_signed_source = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="make_signed_source"
        )
        self.assertQuerySetEqual(
            make_signed_source.dependencies.all(), [confirm]
        )

    def test_populate_make_signed_source_confirmation(self) -> None:
        """If confirming, make_signed_source depends on the confirm task."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_confirmation": True,
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {"amd64": ["hello"]},
                "make_signed_source_key": "ABC123",
            }
        )

        confirm = workflow.children.get(
            task_type=TaskTypes.WAIT, task_name="confirm"
        )
        self.assertEqual(
            confirm.task_data,
            {
                "auto_confirm_if_no_failure": False,
                "confirm_label": "Continue the workflow",
            },
        )
        self.assertEqual(
            confirm.workflow_data_json,
            {"display_name": "Confirm", "step": "confirm", "needs_input": True},
        )
        self.assertQuerySetEqual(confirm.dependencies.all(), [])

        make_signed_source = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="make_signed_source"
        )
        self.assertQuerySetEqual(
            make_signed_source.dependencies.all(), [confirm]
        )

    def test_populate_package_upload(self) -> None:
        """Populate create an upload_package workflow."""
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "architectures": ["amd64", "i386"],
                "enable_upload": True,
                "vendor": "debian",
                "codename": "trixie",
                "upload_merge_uploads": False,
                "upload_since_version": "1.1",
                "upload_target": "ftp://user@example.org/pub/UploadQueue1/",
                "upload_target_distribution": "debian:testing",
            }
        )

        package_upload = workflow.children.get(
            task_name="package_upload", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            package_upload.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "trixie",
                "delayed_days": None,
                "key": None,
                "binary_key": None,
                "merge_uploads": False,
                "require_signature": True,
                "since_version": "1.1",
                "source_artifact": self.source_artifact.id,
                "target": "ftp://user@example.org/pub/UploadQueue1/",
                "target_distribution": "debian:testing",
                "vendor": "debian",
            },
        )

        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": "package upload (hello)",
                "step": "package_upload/hello",
            },
        )
        self.assertQuerySetEqual(package_upload.dependencies.all(), [])

        # PackageUpload.populate() was called and created its tasks.
        # MakeSourcePackageUpload only requires the source artifact and so
        # can start immediately.  Everything else is blocked.
        self.assertQuerySetEqual(
            package_upload.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            # Work requests to construct, sign, and upload the source package.
            [
                (
                    TaskTypes.WORKER,
                    "makesourcepackageupload",
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "makesourcepackageupload",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ]
            # Work requests to sign and upload each of the two binary packages.
            + [
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ]
            * 2,
        )

    def test_populate_package_upload_signed_source(self) -> None:
        """Populate creates package_upload workflows for signed source."""
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "architectures": ["amd64", "i386"],
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {
                    "amd64": ["hello-amd64-signed-template"]
                },
                "make_signed_source_key": "ABC123",
                "enable_upload": True,
                "vendor": "debian",
                "codename": "trixie",
                "upload_merge_uploads": False,
                "upload_since_version": "1.1",
                "upload_target": "ftp://user@example.org/pub/UploadQueue1/",
                "upload_target_distribution": "debian:testing",
            }
        )

        initial_sbuild_workflow = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )
        make_signed_source = workflow.children.get(
            task_name="make_signed_source", task_type=TaskTypes.WORKFLOW
        )
        extract_for_signing = make_signed_source.children.get(
            task_name="extractforsigning", task_type=TaskTypes.WORKER
        )
        sign = make_signed_source.children.get(
            task_name="sign", task_type=TaskTypes.SIGNING
        )
        assemble_signed_source = make_signed_source.children.get(
            task_name="assemblesignedsource", task_type=TaskTypes.WORKER
        )
        package_upload_source, package_upload_signed_amd64 = (
            workflow.children.filter(
                task_name="package_upload", task_type=TaskTypes.WORKFLOW
            ).order_by("id")
        )

        self.assertEqual(
            package_upload_source.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "trixie",
                "delayed_days": None,
                "key": None,
                "binary_key": None,
                "merge_uploads": False,
                "require_signature": True,
                "since_version": "1.1",
                "source_artifact": self.source_artifact.id,
                "target": "ftp://user@example.org/pub/UploadQueue1/",
                "target_distribution": "debian:testing",
                "vendor": "debian",
            },
        )

        self.assertEqual(
            package_upload_source.workflow_data_json,
            {
                "display_name": "package upload (hello)",
                "step": "package_upload/hello",
            },
        )
        self.assertQuerySetEqual(package_upload_source.dependencies.all(), [])

        # PackageUpload.populate() was called and created its tasks.
        # MakeSourcePackageUpload only requires the source artifact and so
        # can start immediately.  Everything else is blocked.
        self.assertQuerySetEqual(
            package_upload_source.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            # Work requests to construct, sign, and upload the source package.
            [
                (
                    TaskTypes.WORKER,
                    "makesourcepackageupload",
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "makesourcepackageupload",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ]
            # Work requests to sign and upload each of the two binary packages.
            + [
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ]
            * 2,
        )

        self.assertEqual(
            package_upload_signed_amd64.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": (
                                "signed-source-amd64-"
                                "hello-amd64-signed-template|build-"
                            ),
                        },
                    }
                ],
                "codename": "trixie",
                "delayed_days": None,
                "key": None,
                "binary_key": None,
                "merge_uploads": False,
                "require_signature": True,
                "since_version": "1.1",
                "source_artifact": (
                    "internal@collections/name:"
                    "signed-source-amd64-hello-amd64-signed-template"
                ),
                "target": "ftp://user@example.org/pub/UploadQueue1/",
                "target_distribution": "debian:testing",
                "vendor": "debian",
            },
        )

        self.assertEqual(
            package_upload_signed_amd64.workflow_data_json,
            {
                "display_name": (
                    "package upload (amd64/hello-amd64-signed-template)"
                ),
                "step": "package_upload/amd64/hello-amd64-signed-template",
            },
        )

        # The workflow can't orchestrate uploads of the signed source
        # package and its binaries until the signed source package has been
        # assembled.
        self.assertEqual(
            package_upload_signed_amd64.status, WorkRequest.Statuses.BLOCKED
        )
        self.assertQuerySetEqual(
            package_upload_signed_amd64.dependencies.all(),
            [assemble_signed_source],
        )
        self.assertQuerySetEqual(package_upload_signed_amd64.children.all(), [])

        # Assembling the signed source package causes the remaining uploads
        # to be orchestrated.
        self.simulate_sbuild_workflow_completion(
            initial_sbuild_workflow, use_data=False
        )
        self.playground.create_signing_input_artifact(
            binary_package_name="hello-signed",
            workspace=extract_for_signing.workspace,
            work_request=extract_for_signing,
        )
        extract_for_signing.refresh_from_db()
        self.playground.advance_work_request(
            extract_for_signing,
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        self.playground.create_signing_output_artifact(
            purpose=KeyPurpose.UEFI,
            fingerprint="ABC123",
            binary_package_name="hello-signed",
            workspace=sign.workspace,
            work_request=sign,
        )
        sign.refresh_from_db()
        self.playground.advance_work_request(
            sign, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        self.playground.create_upload_artifacts(
            src_name="hello-signed-template",
            source=True,
            binary=False,
            workspace=assemble_signed_source.workspace,
            work_request=assemble_signed_source,
        )
        assemble_signed_source.refresh_from_db()
        self.playground.advance_work_request(
            assemble_signed_source,
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        package_upload_signed_amd64.refresh_from_db()
        self.assertEqual(
            package_upload_signed_amd64.status, WorkRequest.Statuses.PENDING
        )
        self.schedule_and_run_workflow(workflow)

        package_upload_signed_amd64.refresh_from_db()
        self.assertEqual(
            package_upload_signed_amd64.status, WorkRequest.Statuses.RUNNING
        )
        self.assertQuerySetEqual(
            package_upload_signed_amd64.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                # Work requests to sign and upload the source package.
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
                # Work requests to sign and upload the binary package.
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ],
        )

    def test_populate_upload_package_no_include_source_binaries(self) -> None:
        """Populate create upload_workflow without source neither binaries."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_upload": True,
                "vendor": "debian",
                "codename": "trixie",
                "upload_merge_uploads": False,
                "upload_since_version": "1.1",
                "upload_target": "ftp://user@example.org/pub/UploadQueue1/",
                "upload_include_source": False,
                "upload_include_binaries": False,
            }
        )

        package_upload = workflow.children.get(
            task_name="package_upload", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            package_upload.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": [],
                "codename": "trixie",
                "delayed_days": None,
                "key": None,
                "binary_key": None,
                "merge_uploads": False,
                "require_signature": True,
                "since_version": "1.1",
                "source_artifact": None,
                "target": "ftp://user@example.org/pub/UploadQueue1/",
                "target_distribution": None,
                "vendor": "debian",
            },
        )

        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": "package upload (hello)",
                "step": "package_upload/hello",
            },
        )

    def test_populate_package_upload_delayed(self) -> None:
        """Populate an upload_package workflow for a delayed upload."""
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "architectures": ["amd64", "i386"],
                "enable_upload": True,
                "vendor": "debian",
                "codename": "trixie",
                "upload_merge_uploads": False,
                "upload_since_version": "1.1",
                "upload_target": "ftp://user@example.org/pub/UploadQueue1/",
                "upload_target_distribution": "debian:testing",
                "upload_delayed_days": 3,
            }
        )

        package_upload = workflow.children.get(
            task_name="package_upload", task_type=TaskTypes.WORKFLOW
        )

        self.assertEqual(
            package_upload.task_data,
            {
                "arch_all_build_architecture": "amd64",
                "binary_artifacts": [
                    {
                        "collection": "internal@collections",
                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                        "name_matcher": {
                            "kind": CollectionItemMatcherKind.STARTSWITH,
                            "value": "build-",
                        },
                    }
                ],
                "codename": "trixie",
                "delayed_days": 3,
                "key": None,
                "binary_key": None,
                "merge_uploads": False,
                "require_signature": True,
                "since_version": "1.1",
                "source_artifact": self.source_artifact.id,
                "target": "ftp://user@example.org/pub/UploadQueue1/",
                "target_distribution": "debian:testing",
                "vendor": "debian",
            },
        )

        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": "package upload (hello)",
                "step": "package_upload/hello",
            },
        )
        self.assertQuerySetEqual(package_upload.dependencies.all(), [])

        # PackageUpload.populate() was called and created its tasks.
        # MakeSourcePackageUpload only requires the source artifact and so
        # can start immediately.  Everything else is blocked.
        self.assertQuerySetEqual(
            package_upload.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.WORKER,
                    "makesourcepackageupload",
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "makesourcepackageupload",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ]
            + [
                (
                    TaskTypes.WAIT,
                    "externaldebsign",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WORKER,
                    "sbuild",
                ),
                (
                    TaskTypes.SERVER,
                    "packageupload",
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.WAIT,
                    "externaldebsign",
                ),
            ]
            * 2,
        )

    def test_populate_package_upload_qa_failure_policy_fail(self) -> None:
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_autopkgtest": True,
                "qa_failure_policy": QAFailurePolicy.FAIL,
                "enable_upload": True,
            }
        )

        self.assertFalse(
            workflow.children.filter(
                task_type=TaskTypes.WAIT, task_name="confirm"
            ).exists()
        )

        qa = workflow.children.get(task_type=TaskTypes.WORKFLOW, task_name="qa")
        package_upload = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_upload"
        )
        self.assertQuerySetEqual(package_upload.dependencies.all(), [qa])

    def test_populate_package_upload_qa_failure_policy_confirm(self) -> None:
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_autopkgtest": True,
                "qa_failure_policy": QAFailurePolicy.CONFIRM,
                "enable_upload": True,
            }
        )

        qa = workflow.children.get(task_type=TaskTypes.WORKFLOW, task_name="qa")
        confirm = workflow.children.get(
            task_type=TaskTypes.WAIT, task_name="confirm"
        )
        self.assertEqual(
            confirm.task_data,
            {
                "auto_confirm_if_no_failure": True,
                "confirm_label": "Continue the workflow despite QA failures",
            },
        )
        self.assertEqual(
            confirm.workflow_data_json,
            {"display_name": "Confirm", "step": "confirm", "needs_input": True},
        )
        self.assertQuerySetEqual(confirm.dependencies.all(), [qa])

        package_upload = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_upload"
        )
        self.assertQuerySetEqual(package_upload.dependencies.all(), [confirm])

    def test_populate_package_upload_confirmation(self) -> None:
        """If confirming, package_upload depends on the confirm task."""
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "enable_confirmation": True,
                "enable_upload": True,
            }
        )

        confirm = workflow.children.get(
            task_type=TaskTypes.WAIT, task_name="confirm"
        )
        self.assertEqual(
            confirm.task_data,
            {
                "auto_confirm_if_no_failure": False,
                "confirm_label": "Continue the workflow",
            },
        )
        self.assertEqual(
            confirm.workflow_data_json,
            {"display_name": "Confirm", "step": "confirm", "needs_input": True},
        )
        self.assertQuerySetEqual(confirm.dependencies.all(), [])

        package_upload = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_upload"
        )
        self.assertQuerySetEqual(package_upload.dependencies.all(), [confirm])

    def test_populate_package_publish(self) -> None:
        """``populate`` creates a ``package_publish`` workflow."""
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        architectures = ["amd64", "i386"]
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "suite": f"sid@{CollectionCategory.SUITE}",
                "architectures": architectures,
            }
        )

        expected_binary_artifacts = {
            "collection": "internal@collections",
            "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
            "name_matcher": {
                "kind": CollectionItemMatcherKind.STARTSWITH,
                "value": "build-",
            },
        }
        sbuild_workflow = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )
        package_publish = workflow.children.get(
            task_name="package_publish", task_type=TaskTypes.WORKFLOW
        )
        self.assertEqual(
            package_publish.task_data,
            {
                "binary_artifacts": [expected_binary_artifacts],
                "replace": False,
                "source_artifact": self.source_artifact.id,
                "target_suite": f"sid@{CollectionCategory.SUITE}",
            },
        )
        self.assertEqual(
            package_publish.workflow_data_json,
            {
                "display_name": "publish packages (hello)",
                "step": "package_publish/hello",
            },
        )

        # The workflow can't orchestrate publication until binaries have
        # been built.
        self.assertEqual(package_publish.status, WorkRequest.Statuses.BLOCKED)
        self.assertQuerySetEqual(
            package_publish.dependencies.all(),
            # https://github.com/typeddjango/django-stubs/pull/3087
            sbuild_workflow.children.all(),  # type: ignore[arg-type]
        )
        self.assertQuerySetEqual(package_publish.children.all(), [])

        # Completing the builds causes the publishing workflow to be
        # orchestrated.
        binaries, _ = self.simulate_sbuild_workflow_completion(
            sbuild_workflow, use_data=False
        )
        package_publish.refresh_from_db()
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(workflow)

        # PackagePublish.populate() was called and created its task.
        expected_copies = [
            {
                "source_items": [f"{self.source_artifact.id}@artifacts"],
                "target_collection": sid.id,
                "unembargo": False,
                "replace": False,
                "variables": {"component": "main", "section": "misc"},
            }
        ]
        for arch in architectures:
            expected_copies += [
                {
                    "source_items": [f"{binary_artifact.id}@artifacts"],
                    "target_collection": sid.id,
                    "unembargo": False,
                    "replace": False,
                    "variables": {
                        "component": "main",
                        "section": "misc",
                        "priority": "optional",
                    },
                }
                for binary_artifact in binaries[arch]
            ]
        self.assertQuerySetEqual(
            package_publish.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "task_data",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.SERVER,
                    "copycollectionitems",
                    {"copies": expected_copies},
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.INTERNAL,
                    "workflow",
                    {},
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.SERVER,
                    "copycollectionitems",
                ),
            ],
        )

    def test_populate_package_publish_signed_source(self) -> None:
        """``populate`` publishes signed source."""
        sid = self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        architectures = ["amd64", "i386"]
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "suite": f"sid@{CollectionCategory.SUITE}",
                "architectures": architectures,
                "enable_make_signed_source": True,
                "make_signed_source_purpose": KeyPurpose.UEFI,
                "signing_template_names": {
                    "amd64": ["hello-amd64-signed-template"]
                },
                "make_signed_source_key": "ABC123",
            }
        )

        initial_sbuild_workflow = workflow.children.get(
            task_name="sbuild", task_type=TaskTypes.WORKFLOW
        )
        make_signed_source = workflow.children.get(
            task_name="make_signed_source", task_type=TaskTypes.WORKFLOW
        )
        extract_for_signing = make_signed_source.children.get(
            task_name="extractforsigning", task_type=TaskTypes.WORKER
        )
        sign = make_signed_source.children.get(
            task_name="sign", task_type=TaskTypes.SIGNING
        )
        assemble_signed_source = make_signed_source.children.get(
            task_name="assemblesignedsource", task_type=TaskTypes.WORKER
        )
        package_publish_source, package_publish_signed_amd64 = (
            workflow.children.filter(
                task_name="package_publish", task_type=TaskTypes.WORKFLOW
            ).order_by("id")
        )
        expected_binary_artifacts = {
            "collection": "internal@collections",
            "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
            "name_matcher": {
                "kind": CollectionItemMatcherKind.STARTSWITH,
                "value": "build-",
            },
        }
        self.assertEqual(
            package_publish_source.task_data,
            {
                "binary_artifacts": [expected_binary_artifacts],
                "replace": False,
                "source_artifact": self.source_artifact.id,
                "target_suite": f"sid@{CollectionCategory.SUITE}",
            },
        )
        self.assertEqual(
            package_publish_source.workflow_data_json,
            {
                "display_name": "publish packages (hello)",
                "step": "package_publish/hello",
            },
        )

        # The workflow can't orchestrate publication until binaries have
        # been built.
        self.assertEqual(
            package_publish_source.status, WorkRequest.Statuses.BLOCKED
        )
        self.assertQuerySetEqual(
            package_publish_source.dependencies.all(),
            # https://github.com/typeddjango/django-stubs/pull/3087
            initial_sbuild_workflow.children.all(),  # type: ignore[arg-type]
        )
        self.assertQuerySetEqual(package_publish_source.children.all(), [])

        # Completing the builds causes the publishing workflow to be
        # orchestrated.
        binaries, _ = self.simulate_sbuild_workflow_completion(
            initial_sbuild_workflow, use_data=False
        )
        self.playground.create_signing_input_artifact(
            binary_package_name="hello-signed",
            workspace=extract_for_signing.workspace,
            work_request=extract_for_signing,
        )
        package_publish_source.refresh_from_db()
        self.assertEqual(
            package_publish_source.status, WorkRequest.Statuses.PENDING
        )
        self.schedule_and_run_workflow(workflow)

        # PackagePublish.populate() was called and created its task.
        expected_copies = [
            {
                "source_items": [f"{self.source_artifact.id}@artifacts"],
                "target_collection": sid.id,
                "unembargo": False,
                "replace": False,
                "variables": {"component": "main", "section": "misc"},
            }
        ]
        for arch in architectures:
            for binary_artifact in binaries[arch]:
                copy_item = {
                    "source_items": [f"{binary_artifact.id}@artifacts"],
                    "target_collection": sid.id,
                    "unembargo": False,
                    "replace": False,
                    "variables": {
                        "component": "main",
                        "section": "misc",
                        "priority": "optional",
                    },
                }
                expected_copies.append(copy_item)
        self.assertQuerySetEqual(
            package_publish_source.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "task_data",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.SERVER,
                    "copycollectionitems",
                    {"copies": expected_copies},
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.INTERNAL,
                    "workflow",
                    {},
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.SERVER,
                    "copycollectionitems",
                ),
            ],
        )

        expected_signed_source_artifact = (
            "internal@collections/name:"
            "signed-source-amd64-hello-amd64-signed-template"
        )
        expected_signed_binary_artifacts = {
            "collection": "internal@collections",
            "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
            "name_matcher": {
                "kind": CollectionItemMatcherKind.STARTSWITH,
                "value": (
                    "signed-source-amd64-hello-amd64-signed-template|build-"
                ),
            },
        }
        self.assertEqual(
            package_publish_signed_amd64.task_data,
            {
                "binary_artifacts": [expected_signed_binary_artifacts],
                "replace": False,
                "source_artifact": expected_signed_source_artifact,
                "target_suite": f"sid@{CollectionCategory.SUITE}",
            },
        )

        self.assertEqual(
            package_publish_signed_amd64.workflow_data_json,
            {
                "display_name": (
                    "publish packages (amd64/hello-amd64-signed-template)"
                ),
                "step": "package_publish/amd64/hello-amd64-signed-template",
            },
        )

        # The workflow can't orchestrate publishing the signed source
        # package and its binaries until the signed source package has been
        # assembled.
        package_publish_signed_amd64.refresh_from_db()
        self.assertEqual(
            package_publish_signed_amd64.status, WorkRequest.Statuses.BLOCKED
        )
        self.assertQuerySetEqual(
            package_publish_signed_amd64.dependencies.all(),
            [assemble_signed_source],
        )
        self.assertQuerySetEqual(
            package_publish_signed_amd64.children.all(), []
        )

        # Assembling the signed source package makes progress.
        extract_for_signing.refresh_from_db()
        self.playground.advance_work_request(
            extract_for_signing,
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        self.playground.create_signing_output_artifact(
            purpose=KeyPurpose.UEFI,
            fingerprint="ABC123",
            binary_package_name="hello-signed",
            workspace=sign.workspace,
            work_request=sign,
        )
        sign.refresh_from_db()
        self.playground.advance_work_request(
            sign, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        signed_source = self.playground.create_upload_artifacts(
            src_name="hello-signed-template",
            source=True,
            binary=False,
            section="admin",
            workspace=assemble_signed_source.workspace,
            work_request=assemble_signed_source,
        ).source
        assemble_signed_source.refresh_from_db()
        self.playground.advance_work_request(
            assemble_signed_source,
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        package_publish_signed_amd64.refresh_from_db()
        self.assertEqual(
            package_publish_signed_amd64.status, WorkRequest.Statuses.PENDING
        )
        self.schedule_and_run_workflow(workflow)

        # The workflow needs the binaries as well.
        package_publish_signed_amd64.refresh_from_db()
        self.assertEqual(
            package_publish_signed_amd64.status, WorkRequest.Statuses.BLOCKED
        )
        sbuild_signed_amd64 = make_signed_source.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        sbuild_signed_amd64_task = sbuild_signed_amd64.children.get(
            task_type=TaskTypes.WORKER, task_name="sbuild"
        )
        self.assertQuerySetEqual(
            package_publish_signed_amd64.dependencies.all(),
            [assemble_signed_source, sbuild_signed_amd64_task],
        )

        # Completing the build causes the signed publishing workflow to be
        # orchestrated.
        signed_binaries, _ = self.simulate_sbuild_workflow_completion(
            sbuild_signed_amd64,
            section="admin",
            priority="extra",
            use_data=False,
        )
        package_publish_signed_amd64.refresh_from_db()
        self.assertEqual(
            package_publish_signed_amd64.status, WorkRequest.Statuses.PENDING
        )
        self.schedule_and_run_workflow(workflow)

        expected_signed_copies = [
            {
                "source_items": [f"{signed_source.id}@artifacts"],
                "target_collection": sid.id,
                "unembargo": False,
                "replace": False,
                "variables": {"component": "main", "section": "admin"},
            },
            {
                "source_items": [f"{signed_binaries['amd64'][0].id}@artifacts"],
                "target_collection": sid.id,
                "unembargo": False,
                "replace": False,
                "variables": {
                    "component": "main",
                    "section": "admin",
                    "priority": "extra",
                },
            },
        ]
        self.assertQuerySetEqual(
            package_publish_signed_amd64.children.order_by("id").values_list(
                "task_type",
                "task_name",
                "task_data",
                "status",
                "dependencies__task_type",
                "dependencies__task_name",
            ),
            [
                (
                    TaskTypes.SERVER,
                    "copycollectionitems",
                    {"copies": expected_signed_copies},
                    WorkRequest.Statuses.PENDING,
                    None,
                    None,
                ),
                (
                    TaskTypes.INTERNAL,
                    "workflow",
                    {},
                    WorkRequest.Statuses.BLOCKED,
                    TaskTypes.SERVER,
                    "copycollectionitems",
                ),
            ],
        )

    def test_populate_package_publish_replace(self) -> None:
        """``populate`` can publish with replacement."""
        self.playground.create_collection(
            name="sid",
            category=CollectionCategory.SUITE,
            data={"may_reuse_versions": True},
        )
        workflow = self.orchestrate(
            extra_data={
                "source_artifact": self.source_artifact.id,
                "suite": f"sid@{CollectionCategory.SUITE}",
                "architectures": ["amd64"],
                "publish_replace": True,
            }
        )

        package_publish = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertTrue(package_publish.task_data["replace"])

        source_data = self.source_artifact.create_data()
        assert isinstance(source_data, DebianSourcePackage)
        sbuild_workflow = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        self.simulate_sbuild_workflow_completion(
            sbuild_workflow, use_data=False
        )
        self.schedule_and_run_workflow(workflow)

        copy_collection_items = package_publish.children.get(
            task_type=TaskTypes.SERVER, task_name="copycollectionitems"
        )
        self.assertEqual(
            [
                copy["replace"]
                for copy in copy_collection_items.task_data["copies"]
            ],
            [True, True],
        )

    def test_populate_package_publish_qa_failure_policy_fail(self) -> None:
        self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "suite": f"sid@{CollectionCategory.SUITE}",
                "enable_autopkgtest": True,
                "qa_failure_policy": QAFailurePolicy.FAIL,
            }
        )

        self.assertFalse(
            workflow.children.filter(
                task_type=TaskTypes.WAIT, task_name="confirm"
            ).exists()
        )

        sbuild = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        qa = workflow.children.get(task_type=TaskTypes.WORKFLOW, task_name="qa")
        package_publish = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertQuerySetEqual(
            package_publish.dependencies.all(),
            {*sbuild.children.all(), qa},
            ordered=False,
        )

    def test_populate_package_publish_qa_failure_policy_confirm(self) -> None:
        self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "suite": f"sid@{CollectionCategory.SUITE}",
                "enable_autopkgtest": True,
                "qa_failure_policy": QAFailurePolicy.CONFIRM,
            }
        )

        qa = workflow.children.get(task_type=TaskTypes.WORKFLOW, task_name="qa")
        confirm = workflow.children.get(
            task_type=TaskTypes.WAIT, task_name="confirm"
        )
        self.assertEqual(
            confirm.task_data,
            {
                "auto_confirm_if_no_failure": True,
                "confirm_label": "Continue the workflow despite QA failures",
            },
        )
        self.assertEqual(
            confirm.workflow_data_json,
            {"display_name": "Confirm", "step": "confirm", "needs_input": True},
        )
        self.assertQuerySetEqual(confirm.dependencies.all(), [qa])

        sbuild = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        package_publish = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertQuerySetEqual(
            package_publish.dependencies.all(),
            {*sbuild.children.all(), confirm},
            ordered=False,
        )

    def test_populate_package_publish_confirmation(self) -> None:
        """If confirming, package_publish depends on the confirm task."""
        self.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        workflow = self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": self.source_artifact.id,
                "suite": f"sid@{CollectionCategory.SUITE}",
                "enable_confirmation": True,
            }
        )

        confirm = workflow.children.get(
            task_type=TaskTypes.WAIT, task_name="confirm"
        )
        self.assertEqual(
            confirm.task_data,
            {
                "auto_confirm_if_no_failure": False,
                "confirm_label": "Continue the workflow",
            },
        )
        self.assertEqual(
            confirm.workflow_data_json,
            {"display_name": "Confirm", "step": "confirm", "needs_input": True},
        )
        self.assertQuerySetEqual(confirm.dependencies.all(), [])

        sbuild = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        package_publish = workflow.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertQuerySetEqual(
            package_publish.dependencies.all(),
            {*sbuild.children.all(), confirm},
            ordered=False,
        )

    def test_populate_with_debian_upload(self) -> None:
        """An upload as input will populate."""
        upload_artifacts = self.playground.create_upload_artifacts()
        self.orchestrate(
            extra_data={
                "architectures": ["amd64"],
                "source_artifact": upload_artifacts.upload.id,
            }
        )



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine integration tests.

Test signing service.
"""

import json
import socket
import subprocess
from pathlib import Path
from typing import Any, override
from unittest import TestCase

import requests
import yaml

from debusine.artifacts.models import ArtifactCategory
from debusine.assets.models import SigningKeyData
from utils.client import Client
from utils.common import Configuration
from utils.integration_test_helpers_mixin import IntegrationTestHelpersMixin
from utils.server import DebusineServer


class IntegrationSigningTests(IntegrationTestHelpersMixin, TestCase):
    """
    Integration tests for the signing service.

    These tests assume:
    - debusine-server is running
    - debusine-signing is running (connected to the server)
    - debusine-client is correctly configured
    """

    @override
    def setUp(self) -> None:
        super().setUp()
        # If debusine-server or nginx was launched just before this test,
        # then debusine-server might not be available yet.  Wait for
        # debusine-server to be reachable if it's not ready.
        self.assertTrue(
            DebusineServer.wait_for_server_ready(),
            f"debusine-server should be available (in "
            f"{Configuration.get_base_url()}) before the integration tests "
            f"are run",
        )
        self.signing_input_files: dict[str, bytes] = {}

    @classmethod
    def download_and_extract_binary(
        cls, apt_path: Path, package_name: str, target: Path
    ) -> Path:
        """Download and extract a binary package."""
        subprocess.check_call(
            ["apt-get", "download", package_name],
            cwd=target,
            env=cls.make_apt_environment(apt_path),
        )
        subprocess.check_call(
            [
                "dpkg-deb",
                "-x",
                next(target.glob(f"{package_name}_*.deb")),
                package_name,
            ],
            cwd=target,
        )
        return target / package_name

    @classmethod
    def verify_uefi_signature(
        cls,
        signing_key_data: SigningKeyData,
        image_data: bytes,
        signed_file: dict[str, Any],
    ) -> None:
        """Verify a signed file against its UEFI signing key."""
        with cls._temporary_directory() as temp_path:
            (certificate := temp_path / "uefi.crt").write_bytes(
                signing_key_data.public_key
            )
            (image := temp_path / "image").write_bytes(image_data)
            (signature := temp_path / "image.sig").write_bytes(
                requests.get(signed_file["url"]).content
            )
            subprocess.check_call(
                [
                    "sbverify",
                    "--cert",
                    certificate,
                    "--detached",
                    signature,
                    image,
                ],
                cwd=temp_path,
            )

    def generate_signing_key(self) -> int:
        """Generate a signing key. Return the task ID."""
        result = DebusineServer.execute_command(
            "create_work_request",
            "signing",
            "generatekey",
            stdin=yaml.safe_dump(
                {"purpose": "uefi", "description": "Test key"}
            ),
        )
        self.assertEqual(result.returncode, 0)
        work_request_id = yaml.safe_load(result.stdout)["work_request_id"]
        assert isinstance(work_request_id, int)
        return work_request_id

    def grant_sign_permission(self, asset_id: int) -> None:
        """Grant the test user permission to sign with asset."""
        result = DebusineServer.execute_command(
            "asset",
            "grant_role",
            str(asset_id),
            "--workspace",
            "debusine/System",
            "signer",
            "Admins",
        )
        self.assertEqual(result.returncode, 0)

    def find_created_asset(self, work_request_id: int) -> dict[str, Any]:
        """Return a dict describing the asset created by work_request_id."""
        result = Client.execute_command(
            "asset", "list", "--work-request", str(work_request_id)
        )
        assets = yaml.safe_load(result.stdout)[0]
        assert isinstance(assets, dict)
        return assets

    def grub_signed_package_name(self) -> str | None:
        """Return the name of a signed grub-efi binary package."""
        dpkg_architecture = subprocess.check_output(
            ["dpkg", "--print-architecture"], text=True
        ).strip()
        match dpkg_architecture:
            case "amd64":
                return "grub-efi-amd64-signed"
            case "i386":
                return "grub-efi-ia32-signed"
            case "arm64":
                return "grub-efi-arm64-signed"
            case _:
                return None

    def sign_template_package(
        self, template_package_name: str, fingerprint: str
    ) -> int:
        """Sign a template package. Return the task id."""
        self.signing_input_files = {}
        with (
            self._temporary_directory() as temp_path,
            self.apt_indexes("bookworm") as apt_path,
        ):
            extracted_template = self.download_and_extract_binary(
                apt_path, template_package_name, temp_path
            )
            packages = json.loads(
                (
                    extracted_template
                    / "usr/share/code-signing"
                    / template_package_name
                    / "files.json"
                ).read_text()
            )["packages"]

            for package_name, metadata in packages.items():
                extracted_package = self.download_and_extract_binary(
                    apt_path, package_name, temp_path
                )
                for file in metadata["files"]:
                    self.assertEqual(file["sig_type"], "efi")
                    self.assertFalse(file["file"].startswith("/"))
                    self.signing_input_files[
                        f"{package_name}/{file['file']}"
                    ] = (extracted_package / file["file"]).read_bytes()

            signing_input_id = self.create_artifact_signing_input(
                template_package_name, self.signing_input_files
            )

        result = DebusineServer.execute_command(
            "create_work_request",
            "signing",
            "sign",
            "--created-by",
            "test-user",
            stdin=yaml.safe_dump(
                {
                    "purpose": "uefi",
                    "unsigned": [signing_input_id],
                    "key": fingerprint,
                }
            ),
        )
        self.assertEqual(result.returncode, 0)
        work_request_id = yaml.safe_load(result.stdout)["work_request_id"]
        assert isinstance(work_request_id, int)
        return work_request_id

    def wait_and_assert_result(
        self, work_request_id: int, result: str = "success"
    ) -> None:
        """Wait for a task to complete, assert success."""
        status = Client.wait_for_work_request_completed(work_request_id, result)
        if not status:
            self.print_work_request_debug_logs(work_request_id)
        self.assertTrue(status)

    def assert_successful_sign(
        self, work_request_id: int, signing_key_data: SigningKeyData
    ) -> int:
        """Wait for sign work request, assert success, return output id."""
        work_request = Client.execute_command(
            "work-request", "show", work_request_id
        )

        # The requested files were signed.
        [signing_output_artifact] = [
            artifact
            for artifact in work_request["artifacts"]
            if artifact["category"] == ArtifactCategory.SIGNING_OUTPUT
        ]
        self.assertEqual(signing_output_artifact["data"]["purpose"], "uefi")
        self.assertEqual(
            signing_output_artifact["data"]["fingerprint"],
            signing_key_data.fingerprint,
        )
        self.assertCountEqual(
            signing_output_artifact["data"]["results"],
            [
                {
                    "file": name,
                    "output_file": f"{name}.sig",
                    "error_message": None,
                }
                for name in self.signing_input_files
            ],
        )
        self.assertCountEqual(
            list(signing_output_artifact["files"]),
            [f"{name}.sig" for name in self.signing_input_files],
        )
        for name in self.signing_input_files:
            self.verify_uefi_signature(
                signing_key_data,
                self.signing_input_files[name],
                signing_output_artifact["files"][f"{name}.sig"],
            )
        signing_output_artifact_id = signing_output_artifact["id"]
        assert isinstance(signing_output_artifact_id, int)
        return signing_output_artifact_id

    def assemble_signed_source(
        self, template_package_name: str, signing_output_artifact_id: int
    ) -> int:
        """Assemble a signed source package, return the task."""
        # We need an environment for this.
        self.create_system_images_mmdebstrap(["bookworm"])
        with self.apt_indexes("bookworm") as apt_path:
            template_artifact_id = self.create_artifact_binary(
                apt_path, template_package_name
            )

        work_request_id = Client.execute_command(
            "work-request",
            "create",
            "--yaml",
            "assemblesignedsource",
            stdin=yaml.safe_dump(
                {
                    "environment": "debian/match:codename=bookworm",
                    "template": template_artifact_id,
                    "signed": [signing_output_artifact_id],
                }
            ),
        )["id"]
        assert isinstance(work_request_id, int)
        return work_request_id

    def assert_successful_assembly(
        self, work_request_id: int, signed_package_name: str
    ) -> None:
        """Examine the results of a sign work request, assert signatures."""
        work_request = Client.execute_command(
            "work-request", "show", work_request_id
        )

        # The resulting source package contains all the signed files.
        [source_package_artifact] = [
            artifact
            for artifact in work_request["artifacts"]
            if artifact["category"] == ArtifactCategory.SOURCE_PACKAGE
        ]
        self.assertEqual(
            source_package_artifact["data"]["name"], signed_package_name
        )
        with self._temporary_directory() as temp_path:
            Client.execute_command(
                "artifact",
                "download",
                "--target-directory",
                temp_path,
                source_package_artifact["id"],
            )
            subprocess.check_call(
                [
                    "dpkg-source",
                    "-x",
                    next(temp_path.glob("*.dsc")).name,
                    "unpacked",
                ],
                cwd=temp_path,
            )
            for name in self.signing_input_files:
                self.assertTrue(
                    (
                        temp_path
                        / "unpacked"
                        / "debian"
                        / "signatures"
                        / f"{name}.sig"
                    ).exists(),
                    f"Assembled source package lacks signature of {name}",
                )

    def test_generate_and_sign(self) -> None:
        """Generate a key and sign something with it."""
        generate_task_id = self.generate_signing_key()
        self.wait_and_assert_result(generate_task_id)
        signing_asset = self.find_created_asset(generate_task_id)
        signing_key_data = SigningKeyData.model_validate(signing_asset["data"])
        signed_package_name = self.grub_signed_package_name()
        if signed_package_name is None:
            # Not all architectures have signed grub, just stop here
            return
        template_package_name = f"{signed_package_name}-template"

        # No permission yet, we expect failure
        signing_task_id = self.sign_template_package(
            template_package_name, signing_key_data.fingerprint
        )
        self.wait_and_assert_result(signing_task_id, result="failure")

        # Grant permission, expect success
        self.grant_sign_permission(signing_asset["id"])
        signing_task_id = self.sign_template_package(
            template_package_name, signing_key_data.fingerprint
        )
        self.wait_and_assert_result(signing_task_id)

        signing_output_artifact_id = self.assert_successful_sign(
            signing_task_id, signing_key_data
        )
        assemble_signed_source_task_id = self.assemble_signed_source(
            template_package_name, signing_output_artifact_id
        )
        self.wait_and_assert_result(assemble_signed_source_task_id)
        self.assert_successful_assembly(
            assemble_signed_source_task_id, signed_package_name
        )

    def test_sign_repository_indexes(self) -> None:
        """Generate and sign indexes for a package repository."""
        suite_name = "signing-experiment"
        self.create_suite(suite_name, architectures=["amd64"])
        self.add_suite_to_archive(suite_name)

        workflow_template_result = Client.execute_command(
            "workflow-template",
            "create",
            "update_suites",
            "update-suites",
            stdin=yaml.safe_dump(
                {"runtime_parameters": {"force_basic_indexes": "any"}}
            ),
        )
        assert workflow_template_result.returncode == 0

        workflow_id = Client.execute_command(
            "workflow",
            "start",
            "update-suites",
            "--yaml",
            stdin=yaml.safe_dump({"force_basic_indexes": True}),
        )["id"]
        assert Client.wait_for_work_request_completed(workflow_id, "success")

        # Check that APT is happy with the resulting archive.
        archive_url = f"https://deb.{socket.getfqdn()}/debusine/System"
        with self.apt_indexes(
            suite_name,
            url=archive_url,
            signed_by=self.get_archive_signing_key(),
            architectures=["amd64"],
        ):
            pass

    def test_sign_repository_indexes_private(self) -> None:
        """Generate and sign indexes for a private package repository."""
        DebusineServer.execute_command(
            "workspace", "define", "--private", "debusine/private", "Admins"
        )
        suite_name = "private-signing-experiment"
        self.create_suite(
            suite_name, workspace_name="private", architectures=["amd64"]
        )
        self.add_suite_to_archive(suite_name, workspace_name="private")
        username = "test"
        token_key = Client.execute_command(
            "archive",
            "token",
            "grant",
            "--yaml",
            "--workspace",
            "private",
            "--suite",
            suite_name,
            "--external-username",
            username,
            "viewer",
        )["token_key"]

        workflow_template_result = Client.execute_command(
            "workflow-template",
            "create",
            "--workspace",
            "private",
            "update_suites",
            "update-suites",
            stdin=yaml.safe_dump(
                {"runtime_parameters": {"force_basic_indexes": "any"}}
            ),
        )
        assert workflow_template_result.returncode == 0

        workflow_id = Client.execute_command(
            "workflow",
            "start",
            "--workspace",
            "private",
            "update-suites",
            "--yaml",
            stdin=yaml.safe_dump({"force_basic_indexes": True}),
        )["id"]
        assert Client.wait_for_work_request_completed(workflow_id, "success")

        # Check that APT is happy with the resulting archive.
        archive_url = f"https://deb.{socket.getfqdn()}/debusine/private"
        with self.apt_indexes(
            suite_name,
            url=archive_url,
            signed_by=self.get_archive_signing_key(
                workspace_name="private",
                suite_name=suite_name,
                auth=(username, token_key),
            ),
            architectures=["amd64"],
            auth=(username, token_key),
        ):
            pass



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Types used by lookups.

These are mainly used by :py:mod:`debusine.server.collections.lookup`, but
are defined here to reduce import cycles.
"""

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Any, Literal, assert_never

from debusine.artifacts.models import CollectionCategory
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.models.artifacts import Artifact
from debusine.db.models.collections import Collection, CollectionItem
from debusine.db.models.work_requests import WorkRequest
from debusine.db.models.workspaces import Workspace
from debusine.db.permissioncontext import PermissionUser
from debusine.tasks.models import LookupDict, LookupSingle


@dataclass(frozen=True, kw_only=True)
class LookupResult:
    """
    The result of a collection item lookup.

    None of :py:class:`Artifact`, :py:class:`Collection`, or
    :py:class:`CollectionItem` are entirely suitable return types: a lookup
    that returns a bare item needs to return the collection item, but in the
    case of a single-segment lookup there's no collection item available,
    only an artifact or collection.

    The least bad option seems to be to define a custom return type
    returning all the different things callers might need.
    """

    result_type: LookupResultType
    lookup: LookupSingle | LookupDict
    parent_collection_lookup: LookupSingle | None = None
    collection_item: CollectionItem | None = None
    artifact: Artifact | None = None
    collection: Collection | None = None

    @classmethod
    def create(
        cls,
        result_type: LookupResultType,
        lookup: LookupSingle | LookupDict,
        parent_collection_lookup: LookupSingle | None = None,
        collection_item: CollectionItem | None = None,
        artifact: Artifact | None = None,
        collection: Collection | None = None,
    ) -> "LookupResult":
        """Instantiate the right LookupResult subclass based on result type."""
        match result_type:
            case LookupResultType.ARTIFACT:
                return LookupResultArtifact(
                    lookup=lookup,
                    parent_collection_lookup=parent_collection_lookup,
                    collection_item=collection_item,
                    artifact=artifact,
                )
            case LookupResultType.BARE:
                return LookupResultBare(
                    lookup=lookup,
                    parent_collection_lookup=parent_collection_lookup,
                    collection_item=collection_item,
                )
            case LookupResultType.COLLECTION:
                return LookupResultCollection(
                    lookup=lookup,
                    parent_collection_lookup=parent_collection_lookup,
                    collection_item=collection_item,
                    collection=collection,
                )
            case _ as unreachable:
                assert_never(unreachable)

    def reconstruct(
        self, workflow_root: WorkRequest | None = None
    ) -> LookupSingle:
        """
        Reconstruct a lookup matching a given result.

        If the lookup result came directly from a :py:class:`LookupSingle`, then
        this is straightforward: just return the original lookup.

        The remaining cases are when the lookup result was one of many resulting
        from looking up a :py:class:`LookupDict`.  If the lookup result was from
        an item in a collection, then the returned lookup will resolve to the
        item with that name in that collection (which may not be the same as the
        original, if the item has been replaced; this is useful for promises
        that may be replaced by real artifacts).  Otherwise, it will resolve to
        the same artifact or collection, providing that it still exists.
        """
        if isinstance(self.lookup, LookupSingle):
            return self.lookup
        if (
            self.parent_collection_lookup is not None
            and self.collection_item is not None
        ):
            return (
                f"{self.parent_collection_lookup}/"
                f"name:{self.collection_item.name}"
            )
        elif self.artifact is not None:
            return f"{self.artifact.id}@artifacts"
        else:
            assert self.collection is not None
            if (
                workflow_root is not None
                and self.collection.id == workflow_root.internal_collection_id
            ):
                return "internal@collections"
            else:
                return f"{self.collection.id}@collections"


class LookupResultBare(LookupResult):
    """
    A collection item lookup containing bare data.

    Used to assist type annotations.
    """

    result_type: Literal[LookupResultType.BARE]
    parent_collection_lookup: LookupSingle
    collection_item: CollectionItem
    artifact: None
    collection: None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize with the default result_type."""
        kwargs["result_type"] = LookupResultType.BARE
        return super().__init__(*args, **kwargs)


class LookupResultArtifact(LookupResult):
    """
    A collection item lookup containing an artifact.

    Used to assist type annotations.
    """

    result_type: Literal[LookupResultType.ARTIFACT]
    artifact: Artifact
    collection: None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize with the default result_type."""
        kwargs["result_type"] = LookupResultType.ARTIFACT
        return super().__init__(*args, **kwargs)


class LookupResultCollection(LookupResult):
    """
    A collection item lookup containing a collection.

    Used to assist type annotations.
    """

    result_type: Literal[LookupResultType.COLLECTION]
    artifact: None
    collection: Collection

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize with the default result_type."""
        kwargs["result_type"] = LookupResultType.COLLECTION
        return super().__init__(*args, **kwargs)


class LookupCacheGroup[Lookup, Result](
    dict[
        tuple[
            Lookup,
            Workspace,
            PermissionUser,
            CollectionCategory | None,
            WorkRequest | None,
            LookupChildType,
        ],
        Result,
    ]
):
    """A cache for a particular kind of lookup."""


@dataclass(slots=True)
class LookupCache:
    """A cache of lookups that have been performed in the current context."""

    single: LookupCacheGroup[LookupSingle, LookupResult] = field(
        default_factory=LookupCacheGroup
    )
    dictionary: LookupCacheGroup[LookupDict, Sequence[LookupResult]] = field(
        default_factory=LookupCacheGroup
    )



# Generated by Django 4.2.28 on 2026-06-10 12:00

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0061_external_user_tokens'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionTokenGrant',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'role',
                    models.CharField(
                        choices=[('owner', 'Owner'), ('viewer', 'Viewer')]
                    ),
                ),
                (
                    'created_at',
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    'collection',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='token_grants',
                        to='db.collection',
                    ),
                ),
                (
                    'created_by',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='user_created_%(class)s',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'token',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='collection_grants',
                        to='db.token',
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name='collectiontokengrant',
            constraint=models.UniqueConstraint(
                fields=('token', 'collection', 'role'),
                name='db_collectiontokengrant_unique',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utility methods used in management module."""

import abc
import datetime as dt
import io
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any, ClassVar, IO, cast, override

import rich
import yaml
from django.conf import settings
from django.core.management import CommandError
from django.db.models import QuerySet, Value
from django.db.models.functions import Coalesce
from rich.table import Table

from debusine.assets.models import AssetCategory
from debusine.db.models import Asset, Scope, WorkerPool, Workspace
from debusine.db.models.assets import AssetUsageRole
from debusine.db.models.auth import GroupMembership


@dataclass
class Column:
    """Represent a column in a tabular output."""

    #: Name in YAML
    name: str
    #: Label in tables
    label: str

    def to_data(self, value: Any) -> Any:
        """Return a value for YAML."""
        return value

    def to_rich(self, value: Any) -> str:
        """Return a value for human-readable output."""
        match value:
            case None:
                return "-"
            case dt.datetime():
                return value.isoformat()
            case _:
                return str(value)


@dataclass
class AttrColumn(Column):
    """Column displaying an attribute of an object."""

    #: Attribute to use for value
    attr: str

    @override
    def to_data(self, value: Any) -> Any:
        return getattr(value, self.attr) if value is not None else None

    @override
    def to_rich(self, value: Any) -> str:
        if value is None:
            return "-"
        value = getattr(value, self.attr)
        return super().to_rich(value)


class Printer(abc.ABC):
    """Print tabular values in human or machine readable formats."""

    #: Description of tabular columns
    columns: ClassVar[list[Column]]

    def __init__(self, yaml: bool) -> None:
        """
        Choose the output type.

        :param yaml: use machine-readable YAML
        """
        self.yaml = yaml

    @abc.abstractmethod
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        """Generate rows to display."""

    def print(
        self, qs: QuerySet[Any], file: IO[str] | io.TextIOBase | None = None
    ) -> None:
        """Generate and print output."""
        if self.yaml:
            self.print_yaml(qs, file)
        else:
            self.print_rich(qs, file)

    def print_yaml(
        self, qs: QuerySet[Any], file: IO[str] | io.TextIOBase | None = None
    ) -> None:
        """Generate machine-readable YAML."""
        rows: list[dict[str, Any]] = []
        for row in self.rows(qs):
            rows.append(  # noqa: PERF401
                {
                    col.name: col.to_data(val)
                    for col, val in zip(self.columns, row)
                }
            )
        print(yaml.dump(rows), file=file)

    def print_rich(
        self, qs: QuerySet[Any], file: IO[str] | io.TextIOBase | None = None
    ) -> None:
        """Generate human-readable output."""
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        for col in self.columns:
            table.add_column(col.label)
        for row in self.rows(qs):
            table.add_row(
                *(col.to_rich(val) for col, val in zip(self.columns, row))
            )
        rich.print(table, file=cast(IO[str], file))


class Tokens(Printer):
    """Display a queryset of tokens."""

    columns = [
        Column("hash", 'Token hash (do not copy)'),
        AttrColumn("user", 'User', "username"),
        Column("created", 'Created'),
        Column("expires", 'Expires'),
        Column("enabled", 'Enabled'),
        Column("comment", 'Comment'),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for token in qs.order_by('user__username', 'created_at'):
            yield [
                token.hash,
                token.user,
                token.created_at,
                token.expire_at,
                token.enabled,
                token.comment,
            ]


class Workers(Printer):
    """Display a queryset of workers."""

    columns = [
        Column("name", 'Name'),
        Column("type", 'Type'),
        Column("registered", 'Registered'),
        Column("connected", 'Connected'),
        AttrColumn("hash", 'Token hash (do not copy)', "hash"),
        AttrColumn("enabled", 'Enabled', "enabled"),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for worker in qs.order_by('registered_at'):
            token = worker.token
            yield [
                worker.name,
                worker.worker_type,
                worker.registered_at,
                worker.connected_at,
                token,
                token,
            ]


class Workspaces(Printer):
    """Display a queryset of workspaces."""

    columns = [
        Column("name", 'Name'),
        Column("public", 'Public'),
        Column("expiration", 'Default Expiration Delay (days)'),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for workspace in qs.order_by('id'):
            yield [
                str(workspace),
                workspace.public,
                workspace.default_expiration_delay.days or "Never",
            ]


class ResourceRoles(Printer):
    """Display a queryset of resource roles."""

    columns = [
        Column("group", 'Group'),
        Column("role", 'Role'),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for role in qs.order_by('group__name', 'role'):
            yield [
                role.group.name,
                role.role,
            ]


class WorkRequests(Printer):
    """Display a queryset of work requests."""

    columns = [
        Column("id", "ID"),
        AttrColumn("worker", "Worker", "name"),
        Column("created_at", "Created"),
        Column("started_at", "Started"),
        Column("completed_at", "Completed"),
        Column("status", "Status"),
        Column("result", "Result"),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for work_request in qs.order_by("created_at"):
            yield [
                work_request.id,
                work_request.worker,
                work_request.created_at,
                work_request.started_at,
                work_request.completed_at,
                work_request.status,
                work_request.result,
            ]


class Users(Printer):
    """Display a queryset of users."""

    columns = [
        Column("username", "User"),
        Column("email", "Email"),
        Column("date_joined", "Joined"),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for user in qs.order_by("date_joined"):
            yield [user.username, user.email, user.date_joined]


class Groups(Printer):
    """Display a queryset of groups."""

    columns = [
        Column("scope", "Scope"),
        Column("workspace", "Workspace"),
        Column("group", "Group"),
        Column("users", "Users"),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for group in qs.order_by(
            "scope", Coalesce("workspace__name", Value("")), "name"
        ):
            yield [
                group.scope.name,
                group.workspace.name if group.workspace else None,
                group.name,
                group.users.count(),
            ]


class GroupMembers(Printer):
    """Display a queryset of group members."""

    columns = [
        Column("username", "User"),
        Column("email", "Email"),
        Column("role", 'Role'),
        Column("date_joined", "Joined"),
    ]

    @override
    def rows(self, qs: QuerySet[GroupMembership]) -> Generator[list[Any]]:
        for membership in qs.order_by("user__date_joined"):
            yield [
                membership.user.username,
                membership.user.email,
                membership.role,
                membership.user.date_joined,
            ]


def sort_keys(d: dict[str, Any]) -> dict[str, Any]:
    """Order keys of a dict."""
    return dict(sorted(d.items()))


class NotificationChannels(Printer):
    """Display a queryset of notification channels."""

    columns = [
        Column("name", "Name"),
        Column("method", "Method"),
        Column("data", "Data"),
    ]

    @override
    def rows(self, qs: QuerySet[Any]) -> Generator[list[Any]]:
        for notification_channel in qs.order_by("name"):
            yield [
                notification_channel.name,
                notification_channel.method,
                sort_keys(notification_channel.data),
            ]


class Assets(Printer):
    """Display a queryset of assets."""

    columns = [
        Column("id", "ID"),
        Column("category", "Category"),
        Column("workspace", "Workspace"),
        Column("data", "Data"),
    ]

    @override
    def rows(self, qs: QuerySet[Asset]) -> Generator[list[Any]]:
        for asset in qs.order_by("id"):
            data: dict[str, Any] = {}
            match asset.category:
                case AssetCategory.SIGNING_KEY:
                    data = {
                        "fingerprint": asset.data["fingerprint"],
                        "purpose": asset.data["purpose"],
                    }
                case _:  # pragma: no cover
                    data = asset.data
            yield [asset.id, asset.category, str(asset.workspace), data]


class AssetUsageRoles(Printer):
    """Display a queryset of asset usage roles."""

    columns = [
        Column("workspace", "Workspace"),
        Column("group", "Group"),
        Column("role", "Role"),
    ]

    @override
    def rows(self, qs: QuerySet[AssetUsageRole]) -> Generator[list[Any]]:
        for asset_usage_role in qs.order_by("id"):
            yield [
                asset_usage_role.resource.workspace.name,
                asset_usage_role.group.name,
                asset_usage_role.role,
            ]


class WorkerPools(Printer):
    """Display a queryset of worker_pools."""

    columns = [
        Column("name", "Name"),
        Column("provider_account", "Provider Account"),
        Column("enabled", "Enabled"),
        Column("specifications", "Specifications"),
        Column("limits", "Limits"),
    ]

    @override
    def rows(self, qs: QuerySet[WorkerPool]) -> Generator[list[Any]]:
        for worker_pool in qs.order_by("id"):
            yield [
                worker_pool.name,
                worker_pool.provider_account.data["name"],
                worker_pool.enabled,
                sort_keys(worker_pool.specifications),
                sort_keys(worker_pool.limits),
            ]


def get_scope(scope_name: str) -> Scope:
    """Look up a scope by name."""
    try:
        return Scope.objects.get(name=scope_name)
    except Scope.DoesNotExist:
        raise CommandError(f"Scope {scope_name!r} not found", returncode=3)


def get_workspace(name: str, require_scope: bool = False) -> Workspace:
    """Get a workspace from the user-provided name."""
    if "/" in name:
        scope_name, workspace_name = name.split("/", 1)
    elif require_scope:
        raise CommandError(
            (
                f"scope_workspace {name!r} should be in the form "
                f"'scopename/workspacename'"
            ),
            returncode=3,
        )
    else:
        scope_name = settings.DEBUSINE_DEFAULT_SCOPE
        workspace_name = name

    scope = get_scope(scope_name)
    try:
        return Workspace.objects.get(name=workspace_name, scope=scope)
    except Workspace.DoesNotExist:
        raise CommandError(
            f"Workspace {workspace_name!r} not found in scope {scope.name!r}",
            returncode=3,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli Workflow commands."""

from pathlib import Path
from textwrap import dedent
from typing import Any, override
from unittest import mock

import yaml

from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import CreateWorkflowRequest
from debusine.test.test_utils import create_work_request_response


class StartTests(BaseCliTests):
    """Tests for the CLI `workflow start` command."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data: dict[str, Any] = {"input": {"source_artifact_id": 1}}

    def get_input_task_data_file(self) -> Path:
        """Write self.task_data to a temporary file and return its path."""
        return self.create_temporary_file(
            contents=yaml.safe_dump(self.task_data).encode("utf-8")
        )

    def test_invalid_template_name(self) -> None:
        """CLI fails if the template name is bad."""
        template_name = "template-name"

        self.enterContext(self.patch_sys_stdin_read("{}"))
        cli = self.create_cli(["workflow", "start", template_name])

        with mock.patch.object(
            Debusine,
            "workflow_create",
            autospec=True,
            side_effect=DebusineError(title=f"invalid {template_name}"),
        ):
            exception = self.assertShowsError(cli.execute)

        self.assertDebusineError(
            exception, {"title": f"invalid {template_name}"}
        )

    def assert_create_workflow(
        self, cli_args: list[str], expected_request: CreateWorkflowRequest
    ) -> None:
        """
        Run workflow start with the given arguments, and check its behaviour.

        :param expected_request: request expected to be generated by the command
        """
        cli = self.create_cli(["workflow", "start", *cli_args])

        response = create_work_request_response(
            base_url=self.get_base_url(self.default_server),
            scope=self.servers[self.default_server]["scope"],
            id=11,
            workspace="test",
        )

        with (
            mock.patch.object(
                Debusine,
                "workflow_create",
                return_value=response,
            ) as workflow_create,
            mock.patch(
                "debusine.client.commands.base.Command.feedback"
            ) as feedback,
            mock.patch(
                "debusine.client.commands.work_requests"
                ".WorkRequestModelCommand.show"
            ) as show,
        ):
            stderr, stdout = self.capture_output(cli.execute)

        workflow_create.assert_called_once_with(expected_request)
        feedback.assert_called_once_with("Workflow started:")
        show.assert_called_once_with(response)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")

    def test_specific_workspace(self) -> None:
        """CLI parses the command line and uses --workspace."""
        template_name = "sbuild-amd64-arm64"
        workspace_name = "Testing"
        expected_request = CreateWorkflowRequest(
            template_name=template_name,
            workspace=workspace_name,
            task_data=self.task_data,
        )
        args = [
            template_name,
            "--workspace",
            workspace_name,
            "--data",
            self.get_input_task_data_file().as_posix(),
        ]
        self.assert_create_workflow(args, expected_request)

    def test_success_data_from_file(self) -> None:
        """CLI creates a workflow with data from a file."""
        expected_request = CreateWorkflowRequest(
            template_name="sbuild-amd64-arm64",
            workspace="developers",
            task_data=self.task_data,
        )
        args = [
            expected_request.template_name,
            "--data",
            self.get_input_task_data_file().as_posix(),
        ]
        self.assert_create_workflow(args, expected_request)

    def test_success_data_from_stdin(self) -> None:
        """CLI creates a workflow with data from stdin."""
        expected_request = CreateWorkflowRequest(
            template_name="sbuild-amd64-arm64",
            workspace="developers",
            task_data=self.task_data,
        )
        self.enterContext(
            self.patch_sys_stdin_read(yaml.safe_dump(self.task_data))
        )
        args = ["sbuild-amd64-arm64", "--data", "-"]
        self.assert_create_workflow(args, expected_request)

    def test_data_is_empty(self) -> None:
        """CLI creates a workflow with empty data."""
        empty_file = self.create_temporary_file()
        cli = self.create_cli(
            [
                "workflow",
                "start",
                "sbuild-amd64-arm64",
                "--data",
                str(empty_file),
            ]
        )

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(
            stderr, "Error: data must be a dictionary. It is empty\n"
        )
        self.assertEqual(stdout, "")

    def test_no_template(self) -> None:
        """CLI creates a workflow directly, without a template."""
        task_name = "package_publish"
        self.task_data = {
            "source_artifact": 1,
            "target_suite": "trixie@debian:suite",
        }
        expected_request = CreateWorkflowRequest(
            task_name=task_name,
            workspace="developers",
            task_data=self.task_data,
        )
        args = [
            task_name,
            "--no-template",
            "--data",
            self.get_input_task_data_file().as_posix(),
        ]
        self.assert_create_workflow(args, expected_request)

    def test_yaml_errors_failed(self) -> None:
        """cli.execute() deals with different invalid task_data."""
        workflows = [
            {
                "task_data": dedent(
                    """\
                    test:
                      name: a-name
                        first-name: some first name
                    """
                ),
                "comment": "yaml.safe_load raises ScannerError",
            },
            {
                "task_data": dedent(
                    """\
                    input:
                      source_url: https://example.com
                     )
                    """
                ),
                "comment": "yaml.safe_load raises ParserError",
            },
        ]

        for workflow in workflows:
            task_data = workflow["task_data"]
            with self.subTest(task_data), self.patch_sys_stdin_read(task_data):
                cli = self.create_cli(["workflow", "start", "workflow-name"])
                stderr, _ = self.capture_output(
                    cli.execute, assert_system_exit_code=3
                )

                self.assertRegex(stderr, "^Error parsing YAML:")
                self.assertRegex(stderr, "Fix the YAML data\n$")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Django REST Framework authentication infrastructure for Debusine."""

from typing import Optional, TYPE_CHECKING, Union, override

from django.contrib.auth.models import AnonymousUser
from rest_framework.authentication import BaseAuthentication

if TYPE_CHECKING:
    from rest_framework.request import Request

    from debusine.db.models import Token, User


class DebusineTokenAuthentication(BaseAuthentication):
    """Restframework authentication for Debusine."""

    # The first element of the returned tuple could be
    # Optional["PermissionUser"], but we keep it in its expanded form to
    # emphasize that it must remain compatible with
    # django.http.request.HttpRequest.user even if we were to change the
    # definition of PermissionUser in future.
    @override
    def authenticate(
        self, request: "Request"
    ) -> tuple[Union["User", AnonymousUser, None], Optional["Token"]] | None:
        """Authenticate the request and return a two-tuple of (user, token)."""
        token = getattr(request._request, "_debusine_token")
        if token is None:
            return None
        if token.user:
            return (token.user, token)
        else:
            return (AnonymousUser(), token)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Metric collector for the Redis backend for prometheus values."""

import json
from collections.abc import Iterable, Sequence
from typing import cast, override

from prometheus_client.metrics_core import Metric
from prometheus_client.registry import Collector, CollectorRegistry
from prometheus_client.samples import Sample

from debusine.server.prometheus_client.redis import redis_client
from debusine.server.prometheus_client.values import MULTIPROCESS_MODE_T


class RedisCollector(Collector):
    """Collector for redis mode."""

    def __init__(self, registry: CollectorRegistry | None) -> None:
        """Initialize a Redis collector and install into registry."""
        self._client = redis_client()
        if registry:
            registry.register(self)

    def _iter_values(self) -> Iterable[tuple[bytes, str]]:
        """Iterate through (key, value) pairs from our redis DB."""
        cursor = 0
        while True:
            # Type bug: https://github.com/redis/redis-py/issues/2399
            cursor, keys = cast(
                tuple[int, Sequence[str]],
                self._client.scan(cursor=cursor, match="value:*"),
            )
            values = cast(Sequence[str], self._client.mget(keys))
            yield from zip(keys, values)
            if cursor == 0:
                break

    @override
    def collect(self) -> Iterable[Metric]:
        """Collect all metrics from our redis DB."""
        metrics: dict[str, Metric] = {}
        histograms: set[str] = set()
        multiprocess: dict[str, MULTIPROCESS_MODE_T] = {}

        for key, value_s in self._iter_values():
            # FIXME: Catch ValueError here, just in case?
            prefix_b, typ_b, multiprocess_mode_b, mmap_key = key.split(b":", 3)
            assert prefix_b == b"value"
            value = float(value_s)

            metric_name, name, labels, help_text = json.loads(mmap_key)

            metric = metrics.get(metric_name)
            if metric is None:
                typ = typ_b.decode()
                metric = Metric(metric_name, help_text, typ)
                metrics[metric_name] = metric

                if typ in ("histogram", "gaugehistogram"):
                    histograms.add(metric_name)

                multiprocess_mode = cast(
                    MULTIPROCESS_MODE_T, multiprocess_mode_b.decode()
                )
                if typ in ("gauge", "gaugehistogram") and multiprocess_mode:
                    multiprocess[metric_name] = multiprocess_mode

            metric.add_sample(name, labels, value)

        for name, multiprocess_mode in multiprocess.items():
            self._accumulate_multiprocess(metrics[name], multiprocess_mode)

        for name in histograms:
            self._fix_histogram(metrics[name])

        return metrics.values()

    def _accumulate_multiprocess(
        self, metric: Metric, multiprocess_mode: MULTIPROCESS_MODE_T
    ) -> None:
        """Merge metrics from multiple processes using multiprocess_mode."""
        # We deal with live/dead with Redis expiry
        if multiprocess_mode.startswith("live"):
            multiprocess_mode = cast(
                MULTIPROCESS_MODE_T, multiprocess_mode[len("live") :]
            )
        if multiprocess_mode == "all":
            return

        by_label: dict[tuple[tuple[str, ...], str], Sample] = {}

        for sample in metric.samples:
            labels = sample.labels.copy()
            labels.pop("pid")
            key = (tuple(labels.values()), sample.name)
            value = sample.value
            if key in by_label:
                current_value = by_label[key].value
                if multiprocess_mode == "min" and value > current_value:
                    continue
                if multiprocess_mode == "max" and value < current_value:
                    continue
                if multiprocess_mode == "sum":
                    value += current_value
                if multiprocess_mode == "mostrecent":
                    raise NotImplementedError(
                        "The 'mostrecent' modes are not supported in "
                        "RedisCollector"
                    )
            by_label[key] = Sample(sample.name, labels, value)

        metric.samples = list(by_label.values())

    def _fix_histogram(self, metric: Metric) -> None:
        """
        Fix-up histogram samples.

        Sort the buckets as expected by a client, and accumulate the values.
        The Histogram class is optimized to only increment the bucket that a
        value first appears in, not larger ones that would also contain it.
        """
        by_label: dict[tuple[tuple[str, ...], str], list[Sample]] = {}

        # Organize into lists of samples by label
        for sample in metric.samples:
            if "le" in sample.labels:
                labels_without_le = sample.labels.copy()
                labels_without_le.pop("le")
                key = (tuple(labels_without_le.values()), sample.name)
            else:
                key = (tuple(sample.labels.values()), sample.name)
            by_label.setdefault(key, []).append(sample)

        metric.samples = []

        for (_labels, name), samples in sorted(by_label.items()):
            if name.endswith("_bucket"):
                # Sort buckets within each label
                samples.sort(key=lambda sample: float(sample.labels["le"]))

                # Accumulate values into larger buckets
                value = 0.0
                for sample in samples:
                    value += sample.value
                    metric.samples.append(
                        Sample(sample.name, sample.labels, value)
                    )

                labels_without_le = sample.labels.copy()
                labels_without_le.pop("le")
                metric.samples.append(
                    Sample(f"{metric.name}_count", labels_without_le, value)
                )

            else:
                metric.samples.extend(samples)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Table column definition and rendering."""

from collections.abc import Iterator
from functools import cached_property
from typing import Any, TYPE_CHECKING, override

from django.db.models import Model
from django.utils.safestring import SafeString

from debusine.web.views.base import TemplateWidget
from debusine.web.views.table.declarative import Attribute
from debusine.web.views.table.ordering import (
    DateTimeOrdering,
    NumberOrdering,
    OrderBys,
    Ordering,
    StringOrdering,
    TableOrdering,
    render_sort_icon,
)

if TYPE_CHECKING:
    from debusine.web.views.table.table import Table


class Column(Attribute):
    """A table column definition."""

    #: Default Ordering class to use
    ordering_class: type[Ordering] = Ordering

    #: Column title
    title: str
    #: Column description
    help_text: str | None
    #: Ordering
    ordering: Ordering | None

    def __init__(
        self,
        title: str,
        *,
        ordering: Ordering | OrderBys | None = None,
        name: str | None = None,
        help_text: str | None = None,
    ) -> None:
        """Store arguments in the Column object for rendering."""
        super().__init__(name=name)
        from debusine.web.views.table.ordering import Ordering, order_by_tuple

        self.title = title
        self.help_text = help_text

        match ordering:
            case Ordering() | None:
                self.ordering = ordering
            case _:
                self.ordering = self.ordering_class(order_by_tuple(ordering))


class StringColumn(Column):
    """A table column that sorts alphabetically."""

    ordering_class: type[Ordering] = StringOrdering


class NumberColumn(Column):
    """A table column that sorts numerically."""

    ordering_class: type[Ordering] = NumberOrdering


class DateTimeColumn(Column):
    """A table column that sorts as datetime."""

    ordering_class: type[Ordering] = DateTimeOrdering


class BoundColumn[M: Model](TemplateWidget):
    """An instantiated table column."""

    template_name = "web/_table_column_header.html"

    def __init__(self, column: Column, table: "Table[Any]") -> None:
        """Store arguments in the Column object for rendering."""
        self.column = column
        self.table = table
        #: Available sort options
        self.orderings: list[TableOrdering[M]] = []

    @override
    def __str__(self) -> str:
        """Return the column name."""
        return self.name

    @override
    def __repr__(self) -> str:
        """Return a debugging representation."""
        return f"BoundColumn({self.name!r})"

    @property
    def name(self) -> str:
        """Return the column name."""
        return self.column.name

    @property
    def title(self) -> str:
        """Return the column title."""
        return self.column.title

    @cached_property
    def sort_icon(self) -> SafeString:
        """Return the sort icon for the table header."""
        return render_sort_icon(self.orderings)

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["column"] = self
        return ctx


class Columns[M: Model]:
    """All columns for a table."""

    def __init__(self, table: "Table[M]") -> None:
        """Columns for a table."""
        self.table = table
        self.columns: dict[str, BoundColumn[M]] = {}

    def add(self, name: str, column: Column) -> BoundColumn[M]:
        """Add a column definition."""
        bc: BoundColumn[M] = BoundColumn(column, self.table)
        self.columns[name] = bc
        return bc

    def __len__(self) -> int:
        """Return the number of columns."""
        return len(self.columns)

    def __getitem__(self, name: str) -> BoundColumn[M]:
        """Get a column by name."""
        return self.columns[name]

    def __iter__(self) -> Iterator[BoundColumn[M]]:
        """Iterate all columns."""
        return iter(self.columns.values())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Implementation of local file backend."""

import datetime as dt
import filecmp
import logging
import os
import shutil
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path, PurePath
from typing import Self, override

from django.conf import settings

from debusine.db.models import File, FileStore
from debusine.server.file_backend.interface import (
    FileBackendEntryInterface,
    FileBackendInterface,
)
from debusine.server.file_backend.models import LocalFileBackendConfiguration
from debusine.utils import atomic_writer

logger = logging.getLogger(__name__)


class LocalFileBackendEntry(
    FileBackendEntryInterface["LocalFileBackend", PurePath]
):
    """
    An entry in a local file backend.

    Local identifiers are the part of the file's path under the base
    directory.
    """

    @classmethod
    def _create_from_file(
        cls, backend: "LocalFileBackend", fileobj: File
    ) -> Self:
        """Make an entry based on a `File`."""
        hash_hex = fileobj.hash_digest.hex()
        identifier = (
            PurePath(hash_hex[0:2])
            / hash_hex[2:4]
            / hash_hex[4:6]
            / f"{hash_hex}-{fileobj.size}"
        )
        return cls(backend=backend, identifier=identifier)

    @override
    def __str__(self) -> str:
        """Return a string describing this entry."""
        return str(self.get_local_path())

    @override
    def get_local_path(self) -> Path:
        """Return the path to a local copy of the file."""
        return self.backend._base_directory / self._identifier

    @override
    @contextmanager
    def get_temporary_local_path(self) -> Generator[Path]:
        yield self.get_local_path()

    @override
    def get_url(
        self,
        content_type: str | None = None,  # noqa: ARG002, U100
    ) -> None:
        """Return None: no remote URL for a file in LocalFileBackend."""
        return None

    @override
    def get_mtime(self) -> dt.datetime | None:
        try:
            return dt.datetime.fromtimestamp(
                self.get_local_path().stat().st_mtime,
                # In general POSIX timestamps are inherently timezone-naive, but
                # we can reasonably assume that a debusine store is on a
                # sufficiently Unix-like file system to be UTC.
                tz=dt.UTC,
            )
        except FileNotFoundError:
            return None

    @staticmethod
    def _create_subdirectories(
        base_directory: Path, subdirectory: Path
    ) -> None:
        """
        Create the subdirectory in base_directory.

        :param base_directory: directory (must exist) to create the subdirectory
        :param subdirectory: such as "a/b/c", created in base_directory
        """
        accumulating = base_directory

        for part in subdirectory.parts:
            accumulating = accumulating / Path(part)
            accumulating.mkdir(exist_ok=True)

    @classmethod
    def _sync_file(cls, path: Path) -> None:
        """Flush `path` and its directory to disk."""
        # same approach (open mode) as sync from coreutils

        # coreutils include O_NONBLOCK to avoid blocking if the file
        # is a FIFO. Files in debusine are not FIFOs
        try:
            fd = os.open(path, os.O_RDONLY)
        except OSError as exc:
            logger.debug(  # noqa: G200
                "Could not open %s for flushing (%s)", path, exc
            )
            return

        os.fsync(fd)
        os.close(fd)

        if path.is_file():
            cls._sync_file(path.parent)

    @override
    def add(self, local_path: Path) -> None:
        """Add the contents of `local_path` to the underlying storage."""
        destination_file = self.get_local_path()
        destination_directory = destination_file.parent

        self._create_subdirectories(
            self.backend._base_directory,
            destination_directory.relative_to(self.backend._base_directory),
        )

        with (
            local_path.open(mode="rb") as local_file,
            atomic_writer(
                destination_file, mode="wb", chmod=0o600
            ) as storage_file,
        ):
            shutil.copyfileobj(local_file, storage_file)

        self._sync_file(destination_file)

    @override
    def same_contents(self, local_path: Path) -> bool:
        # This standard library function maintains a cache indexed by the
        # paths and stat signatures (type, size, mtime) of both files.  We
        # currently assume that the `local_path` argument is always a
        # recently-created file, so the cache will never be hit.  If this
        # assumption no longer holds, then we may want to inline something
        # like `filecmp._do_cmp` instead.
        return filecmp.cmp(local_path, self.get_local_path(), shallow=False)

    @override
    def remove(self) -> None:
        file_path = self.get_local_path()
        file_path.unlink(missing_ok=True)


class LocalFileBackend(FileBackendInterface[LocalFileBackendConfiguration]):
    """Local file backend (in settings.DEBUSINE_STORE_DIRECTORY)."""

    def __init__(self, file_store: FileStore) -> None:
        """Initialize LocalFileBackend."""
        super().__init__()
        self.db_store = file_store

        if (base_directory := self.configuration.base_directory) is None:
            if file_store.name == "Default":
                base_directory = settings.DEBUSINE_STORE_DIRECTORY
            else:
                raise RuntimeError(
                    f'LocalFileBackend {file_store.name} configuration '
                    'requires "base_directory" setting'
                )

        self._base_directory = Path(base_directory)

    def base_directory(self) -> Path:
        """Return the base_directory of this backend."""
        return self._base_directory

    @override
    def get_entry(self, fileobj: File) -> LocalFileBackendEntry:
        return LocalFileBackendEntry._create_from_file(self, fileobj)

    @override
    def list_entries(self) -> Generator[LocalFileBackendEntry]:
        for root, _, filenames in self._base_directory.walk():
            reldirpath = root.relative_to(self._base_directory)
            for filename in sorted(filenames):
                yield LocalFileBackendEntry(
                    backend=self, identifier=reldirpath / filename
                )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Classes and functions supporting database tests."""

import hashlib
import threading
from collections.abc import Callable
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any

from django.db import connection, connections, transaction

from debusine.db.models import File


class RunInParallelTransaction[R: Any]:
    """
    Execute a callable in a separate transaction that gets started in a thread.

    The :py:meth:`start_transaction` starts a new thread in which it starts
    the transaction and executes the callable. The transaction then stays
    open and the thread waits until we decide to call
    :py:meth:`stop_transaction`.

    Alternatively, use :py:meth:`run_and_wait` to do both the above operations
    in a single step.
    """

    def __init__(self, to_run: Callable[[], R]) -> None:
        """
        Create a RunInParallelTransaction object.

        :param to_run: the function/callable object that gets executed in
          a transaction started in a thread.
        """
        self._to_run = to_run
        self._transaction_ready = threading.Event()
        self._transaction_close = threading.Event()
        self._executor: ThreadPoolExecutor | None = None
        self._future: Future[R] | None = None

    def start_transaction(self) -> None:
        """
        Start a transaction and execute the registered callable.

        It does this in a new thread that then waits until
        :py:meth:stop_transaction() gets called.
        """

        def run_in_thread() -> R:
            try:
                with transaction.atomic():
                    result = self._to_run()
                    self._transaction_ready.set()
                    self._transaction_close.wait()
                return result
            finally:
                # Close the DB connection that was opened in this thread
                # (otherwise the connection is left opened and the test
                # runner might not be able to destroy the debusine-test DB)
                connections.close_all()

        self._executor = ThreadPoolExecutor()
        self._future = self._executor.submit(run_in_thread)
        self._transaction_ready.wait()

    def stop_transaction(self) -> R:
        """Close the open transaction and let the thread exit."""
        assert self._executor is not None
        assert self._future is not None
        self._transaction_close.set()
        result = self._future.result()
        self._executor.shutdown(wait=True)
        return result

    def run_and_wait(self) -> R:
        """Run the callable in a transaction and wait for it to finish."""
        self.start_transaction()
        return self.stop_transaction()


def _calculate_hash_from_data(data: bytes) -> bytes:
    hasher = hashlib.new(File.current_hash_algorithm)
    hasher.update(data)
    return hasher.digest()


def raise_db_exception(condition_name: str) -> None:
    """Tell PostgreSQL to raise an exception, aborting the transaction."""
    with connection.cursor() as cursor:
        cursor.execute(f"DO $$ BEGIN RAISE {condition_name}; END; $$")
